#!/usr/bin/env python

"""Apply penalized likelihood to get an ultrametric tree.

References
----------
...
"""

from typing import Dict, Optional, TypeVar
import subprocess
from loguru import logger
import toyplot
import numpy as np
from scipy import stats
from scipy.optimize import minimize, Bounds, LinearConstraint
from scipy.special import factorial
import toytree
from toytree.utils import ToytreeError

logger = logger.bind(name="toytree")

ToyTree = TypeVar("ToyTree")
AGES_MIN = 1e-8
AGES_MAX = 1e8
RATES_MIN = 1e-9
RATES_MAX = 1e5


class Chronos:
    """Return an ultrametric tree using penalized likelihood.

    Scales edge lengths of a non-ultrametric tree to make it 
    ultrametric (all tips align at zero) by treating edge lengths
    as rate x time (which is typically true for ML estimated trees),
    and estimating rate variation among all edges so that the age of 
    tip nodes=0. 

    Parameters
    -----------
    tree: ToyTree
        A toytree instance.
    model: str
        A model of rate variation among edges ("relaxed", ...). 
    weight: float
        Weighting of the rate variation (penalty component) to the 
        penalized likelihood score. This is referred to as lambda 
        by Sanderson. Lower values allow less variation.
    epsilon: float
        Convergence diagnostic... Default=1e-8.
    tol: float
        Tolerance for convergence, the minimum on rates. Default=1e-8.
    verbose: bool
        Prints information on model fitting and results. Default=False.

    Returns
    -------
    ToyTree: with transformed dist (edge length) feature values, and
    features for ages and rates.

    References
    ----------
    Sanderson
    https://github.com/cran/ape/blob/master/R/chronos.R
    Paradis article ...
    """
    def __init__(
        self, 
        tree: ToyTree,
        calibrations: Optional[Dict]=None,
        model: str="relaxed",
        weight: float=1, 
        tol: float=1e-8, 
        verbose: bool=False,
        ):

        # store tree data
        self.tree = tree
        self.model = model
        self.weight = weight
        self.tol = tol
        self.epsilon = 1e-8
        self.calibrations = (
            calibrations if calibrations is not None
            else {self.tree.treenode.idx: (1, 1)}
        )

        # simple logger
        self.print = (print if verbose else iter)

        # store tree components
        self.edges = self.tree.get_edges()
        self.dists = self.tree.get_edge_values("dist")
        self.dists_lf = np.log(factorial(self.dists))
        self.edge_paths = [
            (leaf.idx,) + tuple(i.idx for i in leaf.iter_ancestors()) 
            for leaf in self.tree.treenode.get_leaves()
        ]

        # starting ages
        self.ages_est_idxs = np.zeros(self.tree.nnodes)
        self.ages_init = np.zeros(self.tree.nnodes)
        self.ages_lens = np.zeros(self.edges.shape[0])
        self.get_ages_init()

        # starting rates
        self.rates_init = self.dists / self.ages_lens
        
        # get bound conditions on nodes ages and rates
        self.get_bounds()

        # parameters to estimate and results to store 
        self.params = np.concatenate([
            self.rates_init, 
            self.ages_init[self.ages_est_idxs],
        ])
        self.ploglik = None
        self.results = self.params
        self.print('{} parameters to estimate'.format(self.params.size))

        # gradient incidence matrices for relaxed model
        self.mat_a = np.array([self.edges[:, 1] == i for i in self.ages_est_idxs])
        self.mat_d = np.array([self.edges[:, 0] == i for i in self.ages_est_idxs])

        # gradient incidence matrices for correlated model
        self.ni_ = np.bincount(self.edges[:].flatten())[:-1]
        self.mat_n = np.zeros((self.edges.shape[0], self.edges.shape[0]), dtype=bool)
        for nidx in range(self.tree.nnodes):
            node = self.tree[nidx]
            if node.up:
                self.mat_n[nidx, nidx] = True
                ind = [i.idx for i in node.children]
                self.mat_n[nidx][ind] = True

        # fit the model iteratively
        # self.iterative_model_fit()

    def run(self):
        """
        Calls optimization function, calculate PhiiC and applies results
        to nodes of a Toytree.
        """
        # report results
        self.print('init ploglik: {}'.format(-self.objective(self.params)))
        self.optimize()
        self.print('final ploglik: {}'.format(-self.objective(self.results)))

        # get phic
        self.get_phiic()

        # store final converted tree
        self.tree = self.get_transformed_tree()

    def get_ages_init(self):
        """
        Finds starting values for age lens by applying values between 
        contraints. If no calibrations then it sets root to 1, otherwise, 
        it can accommodate node calibrations.
        """
        self.print("setting initial dates")        

        # hard-coded values for finding starting fit
        root_multiplier = 3
        max_attempts = 10
        attempt = 0

        # iterate to get starting values using hard-coded limits
        while 1:
            
            # set ages to 0 (tips) or nan (all others)
            self.ages_init[:] = np.nan
            self.ages_init[:self.tree.ntips] = 0

            # set the ages of calibrations at their midpoint
            for nidx in self.calibrations:
                age = self.calibrations[nidx]
                self.ages_init[nidx] = age[0] + ((age[1] - age[0]) / 2.)

            # set root age, if not already calibrated, using an approximation
            if np.isnan(self.ages_init[self.tree.treenode.idx]):
                self.ages_init[self.tree.treenode.idx] = root_multiplier * max(self.ages_init)

            # indices of the nodes that need age estimates
            self.ages_est_idxs = np.where(np.isnan(self.ages_init))[0]

            # sort edge paths so that younger ages are set before older ones
            sorted_epaths = sorted(self.edge_paths, key=len)
            # sorted_epaths = sorted(self.edge_paths, key=lambda x: (~np.isnan(x[1:])).argmax())
            # for idx in range(len(self.edge_paths)):
                # print([self.ages_init[i] for i in sorted_epaths[idx]], sorted_epaths[idx])
           
            # sort paths by nodes until first non-nan
            # visit edge paths in order, e.g., [(5, 9, 10), (4, 7, 9, 10), ...]
            # for path in self.edge_paths[::-1]:
            for path in sorted_epaths:
                
                # get ages on this path, e.g., [0.0, nan, nan, 1.0]
                path_ages = [self.ages_init[i] for i in path]
                
                # counting down reversed path indices, e.g., [ 3, 2, 1, 0 ]
                for pidx in range(len(path_ages))[::-1]:
                    
                    # if age is nan then split the distance to next non-nan
                    if np.isnan(path_ages[pidx]):

                        # get value spaced between data with nnodes
                        maxage = path_ages[pidx + 1]
                        minidx = (~np.isnan(path_ages[:pidx])).argmax()
                        minage = path_ages[minidx]
                        interval = maxage - minage
                        extent = interval - (interval / (pidx - minidx + 1))

                        # store the new age value
                        self.ages_init[path[pidx]] = extent
            
            # get age edge lengths
            self.ages_lens = (
                self.ages_init[self.edges[:, 0]] - self.ages_init[self.edges[:, 1]]
            )
            
            # check that all edges are positive in length        
            if all(self.ages_lens >= 0):
                break
            attempt += 1
            root_multiplier *= 1.5
            if attempt == max_attempts:
                raise ToytreeError("bad starting values")

        # store initial values for method validation
        self.init_tree = self.get_transformed_tree()

    def get_bounds(self):
        """
        Stores a list of tuples of (min, max) for every parameter made up 
        of the calibration info or default limits.
        """
        # RATES_MIN = max(RATES_MIN, self.tol)
        self.bounds = [(RATES_MIN, RATES_MAX) for i in self.rates_init]
        self.bounds += [(AGES_MIN, AGES_MAX) for i in self.ages_est_idxs]

        # bounds can be further constrained by hard calibrated nodes
        for nidx in self.calibrations:
            cmin, cmax = self.calibrations[nidx]

            # if calibration is a range then add node back to be estimated.
            if cmin != cmax:
                self.bounds += [cmin, cmax]
                self.ages_est_idxs = np.concatenate(
                    [self.ages_est_idxs, [nidx]]
                )

    def optimize(self, rates=True, ages=True):
        """
        Maximum likelihood optimization of the objective function.
        """
        # estimate all rates and all ages except tips and calibrated nodes
        # fit = minimize(
        #     fun=self.objective,
        #     x0=self.params,
        #     args=(rates, ages),
        #     # method="L-BFGS-B",
        #     method="SLSQP",
        #     # method="TNC",
        #     bounds=self.bounds,
        #     jac=self.gradient,
        #     options={
        #         # "maxiter": 100,
        #         # "disp": False,
        #     }
        # )
        fit = minimize(
            fun=self.objective,
            x0=self.params,
            args=(rates, ages),
            method="L-BFGS-B",
            bounds=self.bounds,
            options={
                # "maxiter": 100,
                # "disp": False,
            }
        )        
        self.results = fit.x
        self.ploglik = fit.fun

    def update(self, params, rates, ages):
        """
        Update one or both parameter sets...
        """
        # separate params
        if rates:
            self.rates_hat = params[:self.rates_init.size]
        else:
            self.rates_hat = self.results[:self.rates_init.size]

        if ages:
            self.ages_hat = params[self.rates_init.size:]
        else:
            self.ages_hat = self.results[self.rates_init.size:]

        # impute estimated ages into the full set of ages
        self.ages_init[self.ages_est_idxs] = self.ages_hat

    def objective(self, params, rates=True, ages=True):
        """
        The penalized likelihood function to MINIMIZE
        """
        # update params on self
        self.update(params, rates, ages)

        # component 2
        if self.model == "relaxed":
            penalty = self.lik_penalty_relaxed().sum()
        elif self.model == "correlated":
            penalty = self.lik_penalty_correlated().sum()

        # return the full score
        return -self.lik_poisson() + (self.weight * penalty)

    def lik_poisson(self):
        """
        The poisson function returns higher values for better fit.
        When dealing with count data the Poisson likelihood statistic 
        can be used to measure how well a particular model describes 
        the data, and is especially appropriate when there are just a 
        few counts per bin, because in that case the Least Squares method 
        is inappropriate. 
        """
        # check that edge lengths are still all positive
        age_lens = self.ages_init[self.edges[:, 0]] - self.ages_init[self.edges[:, 1]]
        if any(age_lens < 0):
            return -1e100

        # calculate likelihood
        ratelens = self.rates_hat * age_lens
        loglik = np.sum(
            self.dists * np.log(ratelens) - ratelens - self.dists_lf
        )
        return loglik

    def lik_penalty_relaxed(self):
        """
        Penalty component to likelihood for the relaxed model.
        """
        alpha = self.rates_hat.mean()
        pdens = sorted(stats.gamma.cdf(self.rates_hat, alpha))
        pcdf = ((np.arange(1, self.rates_hat.size + 1) / self.rates_hat.size) - pdens) ** 2
        return pcdf

    def lik_penalty_correlated(self):
        """
        Penalty component to likelihood for the correlated rates model.
        """
        ele2 = np.var([self.rates_hat[i] for i in self.edges[-2:, 1]])
        ele1 = [
            ((self.rates_hat[edge[0]] - self.rates_hat[edge[1]]) ** 2) 
            for edge in self.edges[:-2]
        ]
        return self.weight * (np.sum(ele1) + ele2)

    def gradient(self, params, rates=True, ages=True):
        """
        Combined gradient function for poisson and penalty components
        """
        # 
        self.update(params, rates, ages)

        # get poisson gradient on params
        gradient = self.gradient_poisson()

        # add the effect of rate model on gradients of rate params
        if self.model == "relaxed":
            gradient[:self.rates_hat.size] += self.gradient_penalty_relaxed()

        # return gradient
        return gradient

    def gradient_poisson(self):
        """
        Gradient function for the Poisson component.
        """
        # get edge lengths between node ages
        age_lens = self.ages_init[self.edges[:, 0]] - self.ages_init[self.edges[:, 1]]

        # gradient for the rates
        rates_gr = (self.dists / self.rates_hat) - age_lens

        # gradient for the dates
        inner = (self.dists / age_lens) - self.rates_hat
        dates_gr = np.sum((inner * self.mat_a) - (inner * self.mat_d), axis=1)

        # return as a concatenated list
        return np.concatenate([rates_gr, dates_gr])

    def gradient_penalty_relaxed(self):
        """
        Gradient function for the relaxed model penalty component
        """
        alpha = self.rates_hat.mean()
        a = stats.gamma.pdf(self.rates_hat, alpha)
        b = np.argsort(self.rates_hat) / self.rates_hat.size
        c = stats.gamma.cdf(self.rates_hat, alpha)
        gradient = 2 * a * (b - c)       
        return self.weight * gradient

    def gradient_penalty_correlated(self):
        """
        Gradient function for the correlated rates penalty component
        """
        # gr <- gradient.poisson(rate, node.time)
        #if (all(gr == 0)) return(gr)

        ## contribution of the penalty for the rates:
        # gr[RATE] <- gr[RATE] - lambda * 2 * 

        # (eta_i * rate - sapply(X, function(x) sum(rate[x])))

        # gradient = (
        #     self.ni[i] * self.rates_hat[i]) - 
        #     self.rates_hat[self.matN[i]].sum()
        #  self.weight * 2 * gradient
        
        # the contribution of the root variance term:
        # if (Nbasal == 2) { # the simpler formulae if there's a basal dichotomy
        #     i <- basal[1]
        #     j <- basal[2]
        #     gr[i] <- gr[i] - lambda * (rate[i] - rate[j])
        #     gr[j] <- gr[j] - lambda * (rate[j] - rate[i])
        # } 

        # if not basal dichotomy then support this...
        # else { # the general case
        #     for (i in 1:Nbasal)
        #         j <- basal[i]
        #         gr[j] <- gr[j] -
        #         lambda*2*(rate[j]*(1 - 1/Nbasal) - sum(rate[basal[-i]])/Nbasal)/(Nbasal - 1)
        #    }
        #    gr
        # }

    def get_transformed_tree(self):
        """
        Returns a toytree with new ages (dists) and rates on nodes from the 
        estimated values in self.results.
        """
        # copy and return tree with starting edges
        newtre = self.tree.copy()
        for node in newtre.treenode.traverse("postorder"):
            if node.up:
                node.dist = self.ages_init[node.up.idx] - self.ages_init[node.idx]
                # node.rate = self.params[node.]
        return newtre

    def iterative_model_fit(self):
        """
        Fits the model by ML iteratively, fitting rates then dates, until
        the change in likelihood score is less than epsilon.
        """
        # initial fit: all params
        self.optimize()
        self.print("initial model fit P-loglik: {}".format(self.ploglik))

        # start iterative loop
        tries = 0
        while 1:

            # params and score at start
            curparams = self.results
            curscore = self.ploglik

            # fit rates holding dates constant
            self.optimize(True, False)
            self.print("optimizing rates P-loglik: {}".format(self.ploglik))

            # reject new model fit
            if self.ploglik > curscore:
                self.results = self.params = curparams
                self.ploglik = curscore
                self.print("rejected rates, P-loglik: {}".format(self.ploglik))
            else:
                curscore = self.ploglik
                curparams = self.results

            # fit dates holding rates constant
            self.optimize(False, True)
            self.print("optimizing dates P-loglik: {}".format(self.ploglik))

            # reject new model fit
            if self.ploglik > curscore:
                self.results = self.params = curparams
                self.ploglik = curscore
                self.print("rejected rates, P-loglik: {}".format(self.ploglik))
            else:
                curscore = self.ploglik
                curparams = self.results

            tries += 1
            if tries > 100:
                break
        # 
        self.print()

    def get_phiic(self):

        # get rates
        self._rates = self.params[:self.rates_init.size]

        # 
        if self.model == 'relaxed':
            # alpha = self._rates.mean()
            # pgams = gamma.pdf(self._rates, alpha ** 2, loc=0, scale=1 / alpha)
            alpha = self.rates_hat.mean()
            pdens = sorted(stats.gamma.cdf(self.rates_hat, alpha))
            dx_ = np.linalg.svd([pdens])[1][0]

        # calculate 
        self.phiic = np.sum([
            -2 * self.ploglik,
            2 * len(self.bounds),
            self.weight * dx_,
        ])
        self.print('loglik: {}'.format(self.lik_poisson()))
        self.print('PHIIC: {}'.format(self.phiic))



    # def phi_correlated(edges, r):
    #     ele2 = np.var([r[i] for i in edges[-2:, 1]])
    #     ele1 = [(r[edge[0]] - r[edge[1]]) ** 2 for edge in edges[:-2]]
    #     return np.sum(ele1) + ele2


    # def gradient_ages(self, rate, time):
    #     age[unknown.ages] <- node.time
    #     real.edge.length <- age[e1] - age[e2]

    #     ## gradient for the rates:
    #     gr <- el/rate - real.edge.length

    #     ## gradient for the dates:
    #     tmp <- el/real.edge.length - rate
    #     tmp2 <- tmp[D_ki]
    #     tmp2[is.na(tmp2)] <- 0
    #     gr.dates <- sapply(A_ki, function(x) sum(tmp[x])) - tmp2

    #     c(gr, gr.dates)


    # def gradient_penalty(self, rate, time):
    #     gr <- gradient.poisson(rate, node.time)
    #     #if (all(gr == 0)) return(gr)

    #     ## contribution of the penalty for the rates:
    #     mean.rate <- mean(rate)

    #     ## rank(rate)/Nb.rates is the same than ecdf(rate)(rate) but faster
    #     # gr[RATE] <- gr[RATE] + lambda*2*dgamma(rate, mean.rate)*(rank(rate)/Nb.rates - pgamma(rate, mean.rate))
    #     # gr



RSTRING = """
library(ape)

# load the Python variables
btree <- read.tree(text="{tree}")
min_ages <- c{min_ages}
max_ages <- c{max_ages}
tips <- list{tips}
lamb <- {lamb}
model <- '{model}'

# run the R code 
nodes <- c()
for (tipset in tips) {{
    mrca <- getMRCA(btree, tipset)
    nodes <- append(nodes, mrca)
}}

calib <- data.frame(node=nodes, age.min=min_ages, age.max=max_ages)

ctree <- chronos(
    btree, 
    lambda=lamb,
    model=model,
    calibration=calib,
)
write.tree(ctree)
"""


def convert_to_rvec(tup):
    return "c('" + "','".join(tup) + "')"

def _run_chronos_in_r(tree, lamb=1, model="relaxed", calibrations=None):
    """
    For testing and validating pen.lik. methods
    """
    calibrations = calibrations if calibrations is not None else {}
    if not calibrations:
        calibrations = {tree.treenode.idx: (1, 1)}
    cmin_age = [str(calibrations[i][0]) for i in sorted(calibrations)]
    cmax_age = [str(calibrations[i][1]) for i in sorted(calibrations)]

    data = {
        "tree": tree.write(),
        "min_ages": "(" + ",".join(cmin_age) + ")",
        "max_ages": "(" + ",".join(cmax_age) + ")",
        "tips": "(" + ",".join(
            convert_to_rvec(tree.get_tip_labels(x))
            for x in sorted(calibrations)
        ) + ")" if calibrations else "()",
        "lamb": lamb,
        "model": model,
    }
    rst = RSTRING.format(**data)
    with open("/tmp/chronos.R", 'w') as out:
        out.write(rst)
    out = subprocess.run(
        ["Rscript", "/tmp/chronos.R"], 
        check=True,
        stdout=subprocess.PIPE,
    )
    out = out.stdout.decode()
    ctre = out.strip().split("[1] ")[-1].strip('"')
    return ctre, out



class TreeSampler:
    """
    Class for applying uncorrelated gamma rate transformations to edges
    to create Generative data for testing the Chronos functions.
    """
    def __init__(self, tree, neff=5e5, gentime=1, gamma=3):
        self.tree = tree
        self.neff = neff
        self.gentime = gentime
        self.gamma = gamma
        self.nnodes = self.tree.nnodes

    def plot_gamma_distributed_rates(self, nsamples=10000, bins=30):
        """
        Draws the stat distribution for verification.
        """
        a = (1 / self.gamma) ** 2
        b = 1
        gamma_rates = np.random.gamma(shape=1 / (a*b**2), scale=a*b**2, size=nsamples)

        # generate a distribution of G or Ne values
        NEVAR = gamma_rates * self.neff
        canvas = toyplot.Canvas(width=600, height=250)
        ax0 = canvas.cartesian(grid=(1, 2, 0), xlabel="gamma distributed Ne variation")
        m0 = ax0.bars(np.histogram(NEVAR, bins=bins))

        GVAR = gamma_rates * self.gentime
        ax1 = canvas.cartesian(grid=(1, 2, 1), xlabel="gamma distributed gentime variation")
        m1 = ax1.bars(np.histogram(GVAR, bins=bins))
        return canvas, (ax0, ax1), (m0, m1)
    
    
    def get_tree(self, N=False, G=False, transform=False, seed=None):
        """
        N (bool):
            Sample gamma distributed variation in Ne values across nodes.
        G (bool):
            Sample gamma distributed variation in G values across nodes.
        transform (int):
            Return edge lengths in requested units by selecting integer value:
            (0) time, (1) coalunits, (2) generations.    
        """
        # get a copy of the sptree
        tree = self.tree.copy()
        
        # use random seed
        if seed:
            np.random.seed(seed)
        
        # get gamma rate multipliers
        a = (1 / self.gamma) ** 2
        b = 1
        
        # sample N values
        if N:
            gamma_rates = np.random.gamma(
                shape=1 / (a*b**2), 
                scale=a*b**2, 
                size=self.nnodes,
            )

            # sample Ne values from gamma dist
            nevals = gamma_rates * self.neff
        
            # apply randomly to nodes of the tree
            tree = tree.set_node_values(
                feature="Ne",
                mapping={i: nevals[i] for i in range(tree.nnodes)}
            )
        else:
            tree = tree.set_node_values("Ne", default=self.neff)
        
        if G:
            gamma_rates = np.random.gamma(
                shape=1 / (a*b**2), 
                scale=a*b**2, 
                size=self.nnodes,
            )

            # sample Ne values from gamma dist
            gvals = gamma_rates * self.gentime
        
            # apply randomly to nodes of the tree
            tree = tree.set_node_values(
                feature="g",
                mapping={i: gvals[i] for i in range(tree.nnodes)}
            )
        else:
            tree = tree.set_node_values("g", default=self.gentime)      
        
        # optionally convert edges to coal units
        if transform == 1:
            tree = tree.set_node_values(
                feature="dist",
                mapping={i: j.dist / (j.Ne * 2 * j.g) for i,j in enumerate(tree)}
            )
            
        elif transform == 2:
            tree = tree.set_node_values(
                feature="dist",
                mapping={i: j.dist / j.g for i,j in enumerate(tree)}
            )            
        return tree


if __name__ == "__main__":

    TREE = toytree.rtree.unittree(10, treeheight=1e6)
    TSA = TreeSampler(TREE, neff=5e5, gentime=1, gamma=3)
    TRE = TSA.get_tree()
    print(TRE)



#!/usr/bin/env python

"""Generic function for parsing trees.

Parse flexible input types (filepath, str, Url) to ToyTree class.
"""

from typing import Union, TypeVar, Optional
from pathlib import Path
# from loguru import logger

from toytree.core import ToyTree, Node
# from toytree.core.node import Node
from toytree.utils import ToytreeError
from toytree.io.src.parse import parse_tree

# logger = logger.bind(name="toytree")

# PEP 484 recommend capitalizing alias names
Url = TypeVar("Url")


def tree(
    data: Union[str, Path, Url],
    feature_prefix: str = "&",
    feature_delim: str = ",",
    feature_assignment: str = "=",
    internal_labels: Optional[str] = None,
) -> ToyTree:
    """Return a ToyTree parsed from variable input types and formats.

    Returns a :class:`ToyTree` object from a variety of optional
    input types, including a newick or nexus string, or a filepath or
    Url to a file containing a newick or nexus string. This function
    will try to auto-detect whether internal node labels are names
    or support values. If the newick string contains additional
    metadata as NHX annotations these will be parsed and stored as Node
    features.

    Parameters
    ----------
    data: Union[str, Path, Url]
        One of several allowed input types containing tree data. The
        str type can be a newick string or valid file path; the Path
        type must be a pathlib.Path object; the Url type is any string
        starting with http, from which str data will be fetched.
    feature_prefix: str
        If NHX meta data is present in the newick string enter the
        common prefix contained in each set of square brackets.
        Common options are "", "&", or "&&NHX:". Default="&".
    feature_delim: str
        If NHX meta data is present in the newick string enter the
        delimiter used to separate key-value pairs. Default=",".
    feature_assignment: str
        If NHX meta data is present in the newick string enter the
        assignment operator between key-value pairs. Default="=".
    internal_labels: str or None
        Labels next to internal nodes are often interchangeably used to
        record node names or support values. Enter "name", "support",
        or a different feature name to assign the values to. Default
        is None which mean toytree will infer as 'name' vs 'support'
        based on whether values are str or numeric type.

    Examples
    --------
    >>> tree = toytree.tree("((a,b),c);")
    >>> tree = toytree.tree("((a:10,b:20)A:100,c:30);")
    >>> tree = toytree.tree("/tmp/test.nex")
    >>> tree = toytree.tree("https://eaton-lab.org/data/Cyathophora.tre")
    """
    # load ToyTree from Node, insures detach if Node is not root.
    if isinstance(data, Node):
        treenode = data.copy(detach=True)
        ttree = ToyTree(treenode)
    # load ToyTree from a newick or nexus from str, URL, or filepath
    elif isinstance(data, (str, Path)):
        ttree = parse_tree(
            data,
            feature_prefix=feature_prefix,
            feature_delim=feature_delim,
            feature_assignment=feature_assignment,
            internal_labels=internal_labels,
        )
    # raise an error (to make an empty tree you must enter empty Node)
    else:
        raise ToytreeError(f"Cannot parse input tree data: {data}")
    return ttree


if __name__ == "__main__":

    URI = "https://eaton-lab.org/data/Cyathophora.tre"
    TREE = tree(URI)
    print(TREE)
    print(TREE.get_node_data())

    import toytree
    tree = toytree.tree("((a,b),c),d);")
    tree


#!/usr/bin/env python

"""Test validation of tree drawing style arguments in `toytree.style`

TreeStyle is a DataClass with a serialized repr. It does not validate
values during setting; instead, a validator is run instide of .draw
when the tree is drawn.
"""

import unittest
import numpy as np
import toyplot
import toytree
from toytree.style import (
    TreeStyle,
    validate_style,
    get_color_mapped_values,
    get_color_mapped_feature,
)
# from toytree.utils import ToytreeError

class Devnull(object):
    """Fake stream to test repr"""
    def write(self, *_): pass


class TestValidator(unittest.TestCase):
    def setUp(self):
        self.itree = toytree.rtree.imbtree(10, seed=123, treeheight=10)
        self.btree = toytree.rtree.baltree(10, seed=123, treeheight=10)
        self.utree = toytree.rtree.unittree(10, seed=123, treeheight=10)
        self.trees = [self.itree, self.btree, self.utree]

    def test_node_colormap_style_repr_before_validation(self):
        """Set node colormap on style dict and show repr (json) w/o error."""
        tree = self.itree
        colors = [
            ("dist", "BlueRed"),
            ("dist", toyplot.color.brewer.map("BlueRed")),
            toytree.color.ToyColor((0, 1, 0, 0.5)),
            (0, 1, 0.5, 0.5),
            "red",
            ["blue"] + ["red"] * (tree.nnodes - 1),
        ]
        for color_arg in colors:
            tree.style.node_colors = color_arg
            print(tree.style, file=Devnull)

    def test_node_colormap_support_nan(self):
        """Set nan values to transparent when color mapping."""
        # only set some data, others have nan.
        svals = {11: 0, 12: 50, 13: 75, 14: 100}
        tree = self.itree.set_node_data("support", svals)
        vals = tree.get_node_data("support").values
        colors = toytree.style.get_color_mapped_feature(
            tree=tree,
            feature="support",
            cmap="BlueRed",
            domain_min=0,
            domain_max=10,
        )

        # assert nans were converted to transparent
        pre = np.where(np.isnan(vals))
        post = np.where(colors == toytree.color.ToyColor("transparent"))
        self.assertEqual(list(pre[0]), list(post[0]))

    def test_node_style_opacity_overrule(self):
        """... opacity should overrule orig fill color opacity."""
        tree = self.itree
        tree.style.node_colors = ['red'] + ['blue'] * (tree.nnodes - 1)
        tree.style.node_style.fill_opacity = 0.2
        style = tree.style.copy()
        validate_style(tree, style)
        # print(style.node_colors)

        tree.style.node_colors = ("dist", "BlueRed")
        tree.style.node_style.fill_opacity = 0.2
        # self.assertEqual()        

    # def test_edge_style_opacity_overrule_2(self):
    #     """..."""
    #     tree = self.itree
    #     tree.style.edge_colors = ['red'] + ['blue'] * (tree.nnodes - 1)
    #     tree.style.edge_style.stroke_opacity = 0.2
    #     style = tree.style.copy()
    #     # validate_style(tree, style)
    #     # print(style.node_colors)

    # def test_set_nan_node_labels_empty_string(self):
    #     """np.nan values on node labels should validate to empty strings."""
    #     self.itree.style.node_colors = "red"
        # style = self.itree.style._validate()
        # self.assertEqual(style, dict...)

    # def test_node_size_to_zeros_for_empty_labels(self):
    #     """Do this?"""

    # def test_node_labels_round_floats_on_validate(self):
    #     """..."""
    #     self.btree.treenode.children[0].support = 99.99
    #     self.btree.treenode.children[1].support = 10

    # def test_anchor_shift_legacy_support(self):
    #     """Support -toyplot-anchor-shift as legacy to anchor-shift."""

        # assert warning message is reported on -toyplot-anchor-shift


        # assert tip_labels support for anchor-shift


        # assert node_labels support for anchor-shift


if __name__ == "__main__":

    unittest.main()



#!/usr/bin/env python

"""Get node coordinate layouts from a tree and style arguments.

This class is primarily intended for internal use.

The Layout class will parse the style args to project Nodes into
their proper coordinates for plotting. This is done very quickly
using cached heights, but if fixed-order or fixed-position args are
used then it requires a tree traversal and to make a copy of the
style dict.
"""

from typing import TypeVar, Optional, Sequence
# from enum import Enum
import numpy as np
from loguru import logger
from toytree.utils import ToytreeError
from toytree.style import TreeStyle
from toytree.layout.src.layout_base import BaseLayout

ToyTree = TypeVar("ToyTree")
Node = TypeVar("Node")
logger = logger.bind(name="toytree")

# pylint: disable=too-many-branches, too-many-statements


# this enum not yet used
class InteriorAlgorithm:
    INTERMEDIATE = 0
    CENTERED = 1
    WEIGHTED = 2


class LinearLayout(BaseLayout):
    """Layout for linear drawing styles: "r", "l", "u", "d"

    The interior_node_layout string is optionally entered as a
    number after the linear style, e.g., "r0", "r1", "u2", etc.
    The options implement 'intermediate', 'centered', or 'weighted',
    positioning of internal nodes relative to their descendants or
    tips.
    """
    def __init__(
        self,
        tree: ToyTree,
        style: TreeStyle,
        fixed_order: Optional[Sequence[str]] = None,
        fixed_position: Optional[Sequence[float]] = None,
        interior_algorithm: int = 0
    ):
        self.interior_algorithm = interior_algorithm
        super().__init__(tree, style, fixed_order, fixed_position)

    def run(self) -> None:
        """Fills the .coords array with x, y coordinates.

        mode: int
            0 = intermediate (mean of children; default)
            1 = centered (mean of descendant tips; used in cloud trees)
            2 = weighted (TODO: not implemented)
        """
        # generate new (x, y) linear coordinates
        if bool(self.interior_algorithm) | (self.fixed_order is not None) | (self.fixed_position is not None):
            self.coords = self._get_fixed_order_and_position_coords()
        else:
            self.coords = np.array(list(self.tree._iter_node_coordinates()))

        # update coordinates for 'use_edge_lengths' and orientation.
        self._update_coordinates()

    def _update_coordinates(self) -> None:
        """Set starting values that will be updated by style args.

        For a linear layout this sets the Node and Tip coordinates. It
        starts with layout='d' and then reorients to any linear: 'rlud'
        """
        # Sets internal dists to 1 and extends leaf edges to align at 0.
        if not self.style.use_edge_lengths:
            self._assign_unit_length_edges()

        # re-orient for layout direction: right, left or down.
        if self.style.layout == "d":
            self.angles = np.repeat(-90, self.tree.ntips)
            self.coords[:, 0] += self.style.xbaseline
            self.coords[:, 1] += self.style.ybaseline
            self.tcoords = self.coords[:self.tree.ntips, :].copy()
            if self.style.tip_labels_align:
                self.tcoords[:, 1] = self.style.ybaseline

        elif self.style.layout == "u":
            self.angles = np.repeat(-90, self.tree.ntips)
            self.coords[:, 1] *= -1
            self.coords[:, 0] += self.style.xbaseline
            self.coords[:, 1] += self.style.ybaseline
            self.tcoords = self.coords[:self.tree.ntips, :].copy()
            if self.style.tip_labels_align:
                self.tcoords[:, 1] = self.style.ybaseline

        elif self.style.layout == "l":
            self.angles = np.zeros(self.tree.ntips)
            self.coords = self.coords[:, [1, 0]]
            self.coords[:, 0] += self.style.xbaseline
            self.coords[:, 1] += self.style.ybaseline
            self.tcoords = self.coords[:self.tree.ntips, :].copy()
            if self.style.tip_labels_align:
                self.tcoords[:, 0] = self.style.xbaseline

        else:
            self.angles = np.zeros(self.tree.ntips)
            self.coords = self.coords[:, [1, 0]]
            self.coords[:, 0] *= -1
            self.coords[:, 0] += self.style.xbaseline
            self.coords[:, 1] += self.style.ybaseline
            self.tcoords = self.coords[:self.tree.ntips, :].copy()
            if self.style.tip_labels_align:
                self.tcoords[:, 0] = self.style.xbaseline

    def _assign_unit_length_edges(self) -> None:
        """When use_edge_length=False this sets all dists to unit 1"""
        for node in self.tree:  # .traverse("postorder"):
            if node.is_leaf():
                self.coords[node.idx, 1] = 0
            else:
                cys = [self.coords[child.idx, 1] for child in node.children]
                self.coords[node.idx, 1] = max(cys) + 1

    def _get_fixed_order_and_position_coords(self) -> np.ndarray:
        """Return coords using fixed args. Requires a tree traversal.

        The idx order of the tips is overriden to re-order them
        according an ordered list of names. This is often used to
        visualize discordance among different trees.
        """
        # get user fixed-positions or use the default range of 0-Ntips
        if self.fixed_position is None:
            positions = np.arange(self.tree.ntips)
        else:
            positions = np.array(self.fixed_position)
            assert positions.size == self.tree.ntips, (
                "fixed_position arg must be same len as ntips.")

        # get user fixed-order as the index of tip name strings
        if self.fixed_order is None:
            idxorder = np.arange(self.tree.ntips)
        else:
            idxorder = np.zeros(self.tree.ntips, dtype=int)
            assert len(self.fixed_order) == self.tree.ntips, (
                "fixed order arg must be same len as ntips.")

            # get indices at which user wants Nodes to be displayed
            tipnames = self.tree.get_tip_labels()
            for idx, name in enumerate(self.fixed_order):
                if name not in tipnames:
                    raise ToytreeError(f"name {name} not in tree.")
                idxorder[tipnames.index(name)] = idx

        # return coordinates using the new fixed_order
        coords = []
        for node in self.tree:  # .traverse("idxorder"):
            if node.is_leaf():
                newx = positions[idxorder[node.idx]]
                coords.append((newx, node._height))
            else:
                # set internal node at midpoint between its children
                # centered placement
                if not self.interior_algorithm:
                    newx = sum([
                        coords[min(node.children).idx][0],
                        coords[max(node.children).idx][0],
                    ]) / 2
                    # newx = np.mean([coords[i.idx][0] for i in node.children])
                # intermediate placement
                elif self.interior_algorithm == 1:
                    tips = node.get_leaves()
                    newx = sum([
                        coords[min(tips).idx][0],
                        coords[max(tips).idx][0],
                    ]) / 2
                # weighted
                else:
                    minc = min(node.children)
                    maxc = max(node.children)
                    mind = minc.dist
                    maxd = maxc.dist
                    minx = coords[minc.idx][0]
                    maxx = coords[maxc.idx][0]
                    newx = (((1 / mind) * minx) + ((1 / maxd) * maxx)) / ((1 / mind) + (1 / maxd))

                coords.append((newx, node._height))
        return np.array(coords)


if __name__ == "__main__":

    import toytree
    toytree.set_log_level("INFO")

    a = toytree.Node("a", dist=1.0)
    b = toytree.Node("b", dist=1.5)
    ab = toytree.Node("ab", dist=0.0)
    ab._add_child(a)
    ab._add_child(b)
    ts = toytree.tree(a)
    ll = toytree.layout.LinearLayout(ts, ts.style)

    # tre = toytree.rtree.rtree(5)
    # tre.style.tip_labels_align = True
    # tre.style.xbaseline = 5
    # tre.style.ybaseline = 2.5
    # tre.style.layout = 'u'
    # lay = LinearLayout(tre, tre.style, None, None, interior_algorithm=1)
    # lay.run()
    # print(lay.coords)
    # print(lay.tcoords)
    # print(lay.interior_algorithm)



#!/usr/bin/env python

"""
Penalized likelihood classes
"""

from typing import Dict, Optional, Tuple, List
import subprocess
from dataclasses import dataclass, field, asdict
from loguru import logger
import numpy as np
import scipy.stats as stats
from scipy.optimize import minimize, Bounds, LinearConstraint
from scipy.special import factorial
import toytree


@dataclass
class PenLikBase:
    tree: toytree.ToyTree
    model: str
    weight: float
    epsilon: float
    tol: float
    verbose: bool
    calibrations: Dict[int,Tuple[int,int]]

@dataclass
class PenLik:
    tree: toytree.ToyTree
    model: str="relaxed"
    weight: float=1
    calibrations: Dict[int,Tuple[int,int]]=None
    epsilon: float=field(default=1e-8, repr=False)
    tol: float=field(default=1e-8, repr=False)
    verbose: bool=field(default=False, repr=False)

    # attrs extracted from tree
    edges: np.ndarray = field(init=False, repr=False)
    dists: np.ndarray = field(init=False, repr=False)
    dists_lf: np.ndarray = field(init=False, repr=False)
    edge_paths: List[Tuple] = field(init=False, repr=False)

    def __post_init__(self):
        # super().__init__(**asdict(self))
        if self.calibrations is None:
            self.calibrations = {self.tree.nnodes: (1, 1)}
        self.edges = self.tree.get_edges()
        self.dists = self.tree.get_node_data("dist").values
        self.dists_lf = np.log(factorial(self.dists))
        self.edge_paths = [
            (leaf.idx,) + tuple(i.idx for i in leaf.iter_ancestors()) 
            for leaf in self.tree.treenode.get_leaves()
        ]


if __name__ == "__main__":
    tree = toytree.rtree.rtree(10)    
    pen = PenLik(tree)
    print(pen)



#!/usr/bin/env python

"""Testing simple alternative validation

Problem
-------
[x] function calls from Tree
[x] default/object style from Tree
[x] Annotate module only has nnodes, etc.
[ ] communicate node_colors(single) -> node_style['fill']
[ ] all substyles as dicts w/ keys converted to css style: (fill-opacity, -toyplot-anchor-shift)
"""

from typing import Union, Dict, TypeVar
import numpy as np
from toytree.color import ToyColor

from toytree.style.src.style_base import NodeStyle

ToyTree = TypeVar("ToyTree")
Color = TypeVar("Color", str, tuple, np.ndarray)

__all__ = [
    "validate_node_style",
]


def validate_node_style(
    tree: ToyTree,
    style: NodeStyle = None,
    **kwargs,
) -> Union[NodeStyle, Dict]:
    """Ensure tip labels are in px sizes and fill color as ToyColor.

    Parameters
    ----------
    style: NodeStyle or None
        A NodeStyle object if validating a ToyTree, else None to get a
        default NodeStyle for validating an annotation.
    **kwargs:
        Any arguments accepted by node_style.
    """
    if not isinstance(style, NodeStyle):
        style = NodeStyle()

    # ToyTree drawing validation modifies values in NodeStyle dict.
    # update default style with user style args
    for key, val in kwargs.items():
        style[key] = val

    # convert stroke to ToyColor
    style.stroke = ToyColor(style.stroke)

    # convert fill to ToyColor
    style.fill = ToyColor(style.fill)

    # require opacity settings even though they are None by default
    if style.stroke_opacity is None:
        style.stroke_opacity = 1.0
    if style.fill_opacity is None:
        style.fill_opacity = 1.0

    # return as a NodeStyle object
    return style


if __name__ == "__main__":

    import toytree

    tree = toytree.rtree.unittree(10, seed=123)



#!/usr/bin/env python

"""Trait evolution subpackage.

"""

from toytree.pcm.src.traits.discrete_markov_model_sim import (
    get_markov_model,
    simulate_discrete_data,
)

from toytree.pcm.src.traits.continuous_brownian_sim import (
    simulate_continuous_brownian,
)



#!/usr/bin/env python

"""Dist metrics for collections of trees.

Can return values for random samples of trees or all pairwise.
"""

import toytree
import numpy as np
import pandas as pd


class Sample:
    def __init__(self, ntrees, method):
        # store input
        self.ntrees = ntrees
        self.method = method

        # store output
        self.pairwisesamplingout = np.zeros((self.ntrees, 1), dtype=int)
        self.randomsamplingout = np.zeros((self.ntrees, 1), dtype=int)
    
    def sampling(self):
        """
        Generate pairwise order/pairings to calculate phylogenetic tree distances 
        (ex. 10 trees provided, compare tree 1 & 2, compare 2 & 3 etc.)
    
        OR Generate random order/pairings to calculate phylogenetic tree distances 
        (ex. 10 trees provided, compare tree 1 & 5, compare 5 & 6 etc.)

        Consensus option is for users to calculate distance metrics between each tree
        relative to the consensus tree
        """
        if self.method == "pairwise":
            self.pairwisesamplingout = np.arange(start=0, stop=self.ntrees, step=1)
            return self.pairwisesamplingout

        if self.method == "random":
            self.randomsamplingout = np.random.choice(self.ntrees, size=self.ntrees, replace=False)
            return self.randomsamplingout

        if self.method == "consensus":
            return self.method


if __name__ == "__main__":
    pass


#!/usr/bin/env python

"""Calculate distance matrices from aligned discrete character data.

TODO: use substitution model functions derived from ML code.

Examples
--------
>>> # get DNA alignment matrix
>>> seqs = toytree...
>>>
>>> # get distance matrices
>>> distJC69 = toytree.infer.get_distance_matrix(data, model="JC69")
>>> distF84 = toytree.infer.get_distance_matrix(data, model="F84")
>>> distK2P = toytree.infer.get_distance_matrix(data, model="K2P")

"""

from typing import Union, Optional
import numpy as np
from numpy.typing import ArrayLike
import pandas as pd
from scipy import spatial
import toytree


def get_distance_matrix(
    data: Union[np.ndarray, pd.DataFrame]) -> pd.DataFrame:
    """Return pairwise distances between samples from a data matrix.

    D measured between samples represents the Euclidean distance,
    also termed the Minkowski distance with p=2, and is calculated
    under the hood using `scipy.distance`.

    Note
    ----
    This method does not currently support missing data. Any missing
    values should be imputed before using.

    Data should be quantitative (e.g., trait measurements) or binary,
    however, multi-state characters (e.g., DNA sequences) should use
    the option ...

    Examples
    --------
    >>>
    >>>
    """
    return pd.DataFrame(spatial.distance_matrix(data, data))


def get_expected_distance(
    time: float,
    length: int,
    model: 'MutationModel',
    ) -> pd.DataFrame:
    """Return expected distance between two samples given a subst model.

    """
    raise NotImplementedError("TODO")



####################################################################
##
##
## should these take seqs as input, or diffs???

def get_distance_JC69(seq1: ArrayLike, seq2: ArrayLike) -> float:
    """Return the JC69 distance between two sequences.

    Parameters
    ----------
    seq1: ArrayLike
        A 1-dimensional array of int or str dtype.
    seq2: ArrayLike
        A 1-dimensional array of same size and dtype as seq1.
    """
    return np.sum(seq1 != seq2) / seq1.size


def get_distance_K2P(ntransitions: int, ntransversions: int, length: int) -> float:
    """Return the K2P distance between two sequences...
    """


if __name__ == "__main__":
    pass



#!/usr/bin/env python

"""Get edges as (child, parent) tuples on a tree.

Methods
-------
Get tuples of Nodes on either side of each edge creating a bipartition
>>> tree.iter_edges('idx')                # ((0, 4), (1, 4), (2, 5), ...

Get a table of the edges in a tree as an array or dataframe
>>> tree.get_edges('idx')                 # np.array([(0, 4), ...])
"""

from typing import Iterator, Tuple, Optional, Union
# from loguru import logger
import pandas as pd
import numpy as np
from toytree import Node, ToyTree
from toytree.core.apis import TreeEnumAPI, add_subpackage_method, add_toytree_method

# logger = logger.bind(name="toytree")

__all__ = [
    "iter_edges",
]


@add_toytree_method(ToyTree)
@add_subpackage_method(TreeEnumAPI)
def iter_edges(
    self,
    feature: Optional[str] = None,
    include_root: bool = False,
) -> Iterator[Tuple[Node, Node]]:
    """Generator of (Node, Node) tuples representing edges in tree.

    Given the current tree rooting edges are yielded in idx order
    as (child, parent). The edges set will include edges connected
    to the root Node even if the tree is unrooted. To instead view
    edges as bipartitions of Node sets see `iter_bipartitions`.

    Parameters
    ----------
    feature: str
        An optional feature of a Node returned in place of the Node
        object to represent it. For example, 'name' or 'idx'.
    include_root: bool
        By default the root edge is excluded whether or not the tree
        is rooted. If this arg is set to True this edge will be
        included as (treenode, None) where None is the non-existent
        parent of the treenode. If tree is unrooted this has no effect.

    Example
    -------
    >>> tree = toytree.rtree.unittree(5, seed=123)
    >>> list(iter_edges(feature='idx'))
    >>> # [(0, 5), (1, 5), (2, 6), (3, 7), (4, 7), (5, 6), (6, 8), (7, 8)]
    """
    if feature:
        for node in self[:-1]:
            yield (getattr(node, feature), getattr(node._up, feature))
    else:
        for node in self[:-1]:
            yield (node, node._up)

    # optionally return root edge
    if self.is_rooted() and include_root:
        node = self[-1]
        if feature:
            yield (getattr(node, feature), None)
        else:
            yield (node, None)



@add_toytree_method(ToyTree)
def get_edges(
    self,
    feature: Optional[str] = None,
    df: bool = False,
    include_root: bool = False,
) -> Union[np.ndarray, pd.DataFrame]:
    """Return matrix of (child, parent) edges.

    Given the current tree rooting edges are yielded in idx order
    as (child, parent).

    Parameters
    ----------
    feature: str or None
        Edges can be represented by Node features such as 'idx' or
        'name' (default='idx'). None returns Node objects.
    df: bool
        If True the matrix is returned as a pd.DataFrame rather
        than as a np.ndarray.

    See Also
    --------
    tree.iter_edges
    """
    edges = list(self.iter_edges(feature=feature, include_root=include_root))
    if df:
        return pd.DataFrame(edges, columns=["child", "parent"])
    return np.array(edges)


if __name__ == "__main__":

    import toytree
    tree = toytree.rtree.unittree(10, seed=123)
    edges = list(tree.iter_edges())
    print(edges)
    print(tree.get_edges('idx', df=1))

    # not ideal, but the include_root option is available.
    # print(tree.get_edges('idx', df=1, include_root=True))    
    # print(tree.get_edges('idx', df=0, include_root=True))
    # print(tree.get_edges(include_root=True))    



#!/usr/bin/env python

"""Utilities for visualizing TreeSequences from tskit or SLiM.

The :class:`ToyTreeSequence` class holds the tskit.trees.TreeSequence 
object in its `.tree_sequence` attribute. It is primarily used to 
generate drawings of one or more trees with the option to display 
mutations, MutationTypes, and a chromosome structure. These latter 
options are mostly for SLiM simulated TreeSequences.

Examples
--------
>>> ts = msprime.sim_ancestry()
>>> mts = msprime.sim_mutations()
>>> tts = ToyTreeSequence(tts)

>>> # Draw an individual tree w/ mutations.
>>> tts.draw_tree(idx=0, ...)
>>> tts.draw_tree(site=30, ...)

# Draw the TreeSequence w/ mutations.
>>> tts.draw_tree_sequence(max_trees=10, chromosome=...)
"""

from typing import Optional, Iterable, Union, Dict, TypeVar, Collection, Mapping
from loguru import logger
import numpy as np
from toytree.utils.src.toytree_sequence_drawing import ToyTreeSequenceDrawing
import toytree


logger = logger.bind(name="toytree")

ToyTree = TypeVar("toytree.ToyTree")
TskitTree = TypeVar("tskit.trees")
TreeSequence = TypeVar("tskit.trees.TreeSequence")


class ToyTreeSequence:
    """Return an instance of a ToyTreeSequence.

    ToyTreeSequence objects wrap around tskit.trees.TreeSequence 
    objects to make it easier to extract subsamples from trees, and 
    visualize one or multiple trees as a sequence.

    Parameters
    ----------
    tree_sequence: tskit.trees.TreeSequence
        A tree sequence object from msprime, tskit, SLiM, etc.
    sample: int or Iterable [int]
        The max number of samples per population to include in a tree.
    seed: int
        A random seed used to sample the samples to include in tree.
    name_dict: Mapping[int,str]
        Optionally include a dictionary mapping tree sequence int node 
        IDs to st names. If not provided then tip Nodes will be auto
        renamed from treesequence data as {pop-id}-{node-id}.
    """
    def __init__(
        self,
        tree_sequence: TreeSequence,
        sample: Union[int, Iterable[int], None] = None,
        seed: Optional[int] = None,
        name_dict: Optional[Mapping[int, str]] = None,
    ):
        # TODO: perhaps combine name_dict and 'sample' request method?

        # store user args
        self.rng = np.random.default_rng(seed)
        """: numpy random number generator seeeded."""
        self.npopulations = len(list(tree_sequence.populations()))
        """: number of populations in the TreeSequence."""
        self.sample: Collection = None
        """: the number of haplotypes to sample from the TreeSequence."""
        self.tree_sequence: TreeSequence = None
        """: the TreeSequence object."""
        self.name_dict: Mapping[int, str] = {} if name_dict is None else name_dict
        self._i = 0
        """: iterable counter."""

        # subsample/simplify treesequence to same or smaller nsamples
        if sample is None:
            self.sample = [tree_sequence.sample_size]
            self.tree_sequence = tree_sequence
        elif isinstance(sample, int):
            self.sample = [sample] * self.npopulations
            self.tree_sequence = self._get_subsampled_ts(tree_sequence)
        else:
            self.sample = sample
            self.tree_sequence = self._get_subsampled_ts(tree_sequence)

    def __len__(self):
        return self.tree_sequence.num_trees

    def __iter__(self):
        return self

    def __next__(self):
        try:
            result = self.at_index(self._i)
        except IndexError as err:
            self._i = 0
            raise StopIteration from err
        self._i += 1
        return result

    def __repr__(self):
        return f"<ToyTreeSequence ntrees={len(self)}>"

    def _get_subsampled_ts(self, tree_sequence: TreeSequence) -> TreeSequence:
        """Return a subsampled (simplify'd) tree sequence.

        Samples a maximum of `self.sample` nodes at time=0 from each
        population. This is useful for working with simulations from
        SLiM where nodes of the entire population are often present.
        """
        samps = []
        for pidx, pop in enumerate(tree_sequence.populations()):
            ndt = tree_sequence.tables.nodes
            mask = (ndt.population == pop.id) & (ndt.time == 0) & (ndt.flags == 1)
            if mask.sum():
                arr = np.arange(mask.shape[0])[mask]
                size = min(arr.size, self.sample[pidx])
                samp = self.rng.choice(arr, size=size, replace=False)
                samps += samp.tolist()
        logger.debug(samps)
        sampled_ts = tree_sequence.simplify(samps, keep_input_roots=True)
        return sampled_ts

    def at_index(self, index: int):
        """Return a ToyTree from the indexed order in the tree_sequence."""
        tree = self.tree_sequence.at_index(index)
        return self._get_toytree(tree)

    def at_site(self, pos: int):
        """Return a ToyTree for a position in the tree_sequence."""
        tree = self.tree_sequence.at(pos)
        return self._get_toytree(tree, site=pos)

    def _get_toytree(self, tree: TskitTree, site: Optional[int] = None) -> ToyTree:
        """Return a ToyTree with mutations for a tree selected by site or index.

        It is assumed here that the tree is fully coalesced, returning
        a multitree for multiple trees present is not yet supported...
        """
        # create a psuedo-root that will either be replaced by the root
        # of the tree, if it has a single one, or will be the root to
        # multipe subtrees if they do not coalesce. This root will have
        # the root nodes as children, but they will only have None as up

        # create the root node
        tsidx_dict = {}

        # warn use if multiple roots are present
        if tree.has_multiple_roots:
            logger.warning(f"tree has multiple ({len(tree.roots)}) roots.")
            # a special name to indicate the root is masked ancestors
            pseudo_root = toytree.Node(name="MASKED-ANCESTORS")
            pseudo_root.tsidx = -1
            for tsidx in tree.roots:
                node = toytree.Node(name=tsidx, dist=tree.branch_length(tsidx))
                node.tsidx = tsidx
                # setting disconnected root children causes more problems
                # than its worth. It was easier to just hide the root node
                # during visualizations if desired.
                # node._up = None
                # pseudo_root._children += (node, )
                pseudo_root._add_child(node)
                tsidx_dict[tsidx] = node
                logger.debug(f"adding root: {tsidx}")

                # if root has no children then
        else:
            tsidx = tree.roots[0]
            pseudo_root = toytree.Node(name=tsidx, dist=0)
            pseudo_root.tsidx = tsidx
            tsidx_dict[tsidx] = pseudo_root

        # root_idx = max(pdict.values())
        # idx_dict = {root_idx: toytree.TreeNode(name=root_idx, dist=0)}
        # idx_dict[root_idx].add_feature("tsidx", root_idx)

        # dict of {child: parent, ...} e.g., {0: 5, 1: 8, 2: 7, ...}
        pdict = tree.get_parent_dict()
        logger.debug(pdict)

        # iterate over child, parent items
        for cidx, pidx in pdict.items():
            # if parent already in tsdict get existing Node
            if pidx in tsidx_dict:
                pnode = tsidx_dict[pidx]

            # else create a new Node for it.
            else:
                pnode = toytree.Node(name=str(pidx), dist=tree.branch_length(pidx))
                pnode.tsidx = pidx
                tsidx_dict[pidx] = pnode
                logger.debug(f"adding child: {cidx} to parent: {pidx}")

            # if child node already exists use Node.
            if cidx in tsidx_dict:
                cnode = tsidx_dict[cidx]

            # create a new Node
            else:
                if cidx in self.name_dict:
                    name = self.name_dict[cidx]
                else:
                    pop = self.tree_sequence.tables.nodes.population[cidx]
                    name = f"p{pop}-{cidx}"
                cnode = toytree.Node(
                    name=name,
                    dist=tree.branch_length(cidx),
                )
                cnode.tsidx = cidx
                tsidx_dict[cidx] = cnode
            pnode._add_child(cnode)

        # wrap TreeNode as a toytree
        # return pseudo_root
        ttree = toytree.ToyTree(pseudo_root)  # root_idx])#.ladderize()

        # add mutations as metadata
        if site is not None:
            ttree.mutations = [i for i in tree.mutations() if i.site == site]
        else:
            ttree.mutations = list(tree.mutations())

        # add dict to translate ts idx to toytree idx
        ttree.tsidx_dict = ttree.get_feature_dict("tsidx", None)
        return ttree

    def draw_tree_sequence(
        self,
        start: int = 0,
        max_trees: int = 10,
        height: int = None,
        width: int = None,
        # show_mutations: bool=True,
        scrollable: bool = True,
        **kwargs,
    ):
        # -> Tuple[toyplot.Canvas, toyplot.coordinates.cartesian, :
        """
        Returns a toyplot Canvas, Axes, and list of marks composing 
        a TreeSequenceDrawing with genealogies corresponding to 
        positions along a chromosome.

        Parameters:
        -----------
        """
        # get list of trees.
        ntrees = len(self)
        tmax = start + min(ntrees, max_trees)
        tree_range = range(start, tmax)
        trees = [self.at_index(i) for i in tree_range]
        breaks = self.tree_sequence.breakpoints(True)[start: tmax + 1]
        tsd = ToyTreeSequenceDrawing(
            trees, breaks,
            width=width, height=height,
            scrollable=scrollable,
            **kwargs,
        )
        return tsd.canvas, tsd.axes, tsd.marks

    def draw_tree(
        self,
        site: Optional[int] = None,
        idx: Optional[int] = None,
        mutation_size: Union[int, Iterable[int]] = 8,
        mutation_style: Dict[str, str] = None,
        mutation_color_palette: Iterable['color'] = None,
        show_label=True,
        **kwargs,
    ):
        """Return a toytree drawing of a selected tree site or index.

        You can select a tree by index or site, but not both. The
        selected tree is extracted from the tree_sequence as a
        ToyTree object with relabeled tips showing the population id.
        Mutations are drawn on edges and colored by MutationType, if
        info is present.
        """
        if (site is None) and (idx is None):
            idx = 0
        if (site is not None) and (idx is not None):
            raise ValueError("must enter a 'site' or 'idx' argument, but not both.")
        if site is not None:
            label = f"tree at site {site}"
            tree = self.at_site(site)
        else:
            ivals = list(self.tree_sequence.breakpoints())
            label = f"tree in interval {idx} (sites {ivals[idx]:.0f}-{ivals[idx+1]:.0f})"
            tree = self.at_index(idx)

        # draw the tree with kwargs
        canvas, axes, mark = tree.draw(**kwargs)

        # standard color palette for mtypes, or user-specified one.
        colormap = (
            mutation_color_palette if mutation_color_palette is not None
            else toytree.color.COLORS1)

        # build mutation marker info
        xpos = []
        ypos = []
        titles = []
        colors = []

        # TODO: iterate over .sites() to show positions and describe
        # whether multiple mutations occurred at a position. This
        # needs to be tested for both msprime and slim data types...
        for mut in tree.mutations:

            # get node id and time using the 'tsidx' (tskit node id)
            node = tree.tsidx_dict[mut.node]
            time = mut.time

            # TODO: make optional.
            # do now show mutations on root node.
            if node.is_root():
                continue

            # project mutation points to proper layout type
            if mark.layout == "l":
                ypos.append(node._x)
                xpos.append(time)
            elif mark.layout == "r":
                ypos.append(node._x)
                xpos.append(-time)
            elif mark.layout == "d":
                xpos.append(node._x)
                ypos.append(time)
            elif mark.layout == "u":
                xpos.append(node._x)
                ypos.append(-time)

            # try to extract mtype from metadata if present.
            try:
                mtype = int(mut.metadata['mutation_list'][0]['mutation_type'])
            except Exception:
                mtype = 0

            # store color and title for this point
            color = colormap[mtype]
            colors.append(color)
            title = (
                f"id: {mut.id}\n"
                f"site: {mut.site:.0f}\n"
                f"time: {mut.time:.0f}\n"
                f"mtype: {mtype}")
            titles.append(title)

        # update mutation style dict
        mstyle = {"stroke": "black"}
        if mutation_style is not None:
            mstyle.update(mutation_style)

        # draw the mutations
        mark2 = axes.scatterplot(
            xpos, ypos,
            marker='o',
            size=mutation_size,
            color=colors if colors else None,
            mstyle=mstyle,
            title=titles if titles else None,
        )

        # add axes labels
        if show_label:
            axes.label.text = label
        return canvas, axes, (mark, mark2)


if __name__ == "__main__":

    # EXAMPLE
    import ipcoal
    # import toyplot.browser
    # import pyslim
    toytree.set_log_level("DEBUG")

    # SLIM/shadie example
    import tskit
    TSFILE = "/tmp/test.trees"
    # full ts
    ts = tskit.load(TSFILE)
    # subsampled ts
    tts = ToyTreeSequence(ts, sample=10)
    # get a toytree
    node = tts.at_index(0)
    print(node)
    for child in node.children:
        print(child, child.up)
    # print(node.draw_ascii())
    toytree.ToyTree(node)  # ._draw_browser()
    # tts._get_toytree(tts.at_index(0))

    # ...
    # TRE = toytree.rtree.unittree(ntips=6, treeheight=1e6, seed=123)
    # MOD = ipcoal.Model(tree=TRE, Ne=1e5, recomb=2e-9, store_tree_sequences=True)
    # MOD.sim_loci(1, 10000)

    # # slim_ts = pyslim.load("/tmp/test2.trees")
    # ts = MOD.ts_dict[0]
    # tts = ToyTreeSequence(ts, sample=5, seed=123)  #.at_index(0)
    # tts.draw_tree(width=None, height=300)  #, idxs=None)# scrollable=True)
    # tts = ToyTreeSequence(slim_ts, sample=5, seed=123).at_index(0)

    # TREE = toytree.rtree.unittree(10, treeheight=1e5)
    # MOD = ipcoal.Model(tree=TREE, Ne=1e4, nsamples=1)
    # MOD.sim_loci(5, 2000)

    # # optional: access the tree sequences directly
    # TS = MOD.ts_dict[0]

    # # get a specific ToyTree
    # TOYTS = ToyTreeSequence(TS)

    # # draw the tree sequence
    # # TOYTS.draw_tree(width=None, height=300, idxs=None, scrollable=True)

    # # draw one tree
    # TOY = TOYTS.at_index(0)
    # print(TOY)



#!/usr/bin/env python

"""General phylogenetic comparative methods utilities.

Note
----
Some of these functions may move to other modules as they are
further developed.

References
----------
- Freckleton, R. P., Harvey, P. H. and M. Pagel, M. (2002)
  Phylogenetic analysis and comparative data: a test and review of
  evidence. _American Naturalist_, *160*, 712-726.
- Pagel, M. (1999) Inferring the historical patterns of biological
  evolution. _Nature_, *401*,877-884.
"""

import os
from typing import Union, Callable
from concurrent.futures import ProcessPoolExecutor
import numpy as np
import pandas as pd
import toytree
from toytree import ToyTree, MultiTree
from toytree.core.apis import add_subpackage_method, PhyloCompAPI


def calculate_posterior(
    function: Callable,
    trees: Union[str, ToyTree, MultiTree],
    njobs: int = 1,
    **kwargs,
) -> pd.DataFrame:
    """Return a DataFrame with posterior of (metric) from a tree set.

    Calculations are parallelized and tree(s) can be loaded from a
    flexible set of options. For super large files reading from a
    multi-line newick file is most memory-efficient.

    Parameters
    ----------
    function: Callable
        A function that takes a tree as input and returns a float
        metric as a pandas.Series.
    trees: ToyTree, MultiTree, or newick file
        Flexible options to input/parse one or multiple trees.
    njobs: int
        Distribute N jobs in parallel using ProcessPoolExecutor
    **kwargs: dict
        A dictionary of arguments to the Callable function.

    Examples
    --------
    >>> ...
    """
    # load data and metadata from newick, toytree, or multitree
    if isinstance(trees, list):
        trees = toytree.mtree(trees)
    if isinstance(trees, str) and os.path.exists(trees):
        with open(trees) as tree_generator:
            tre = toytree.tree(next(tree_generator))
            ntips = tre.ntips
            ntrees = sum(1 for i in tree_generator) + 1
            tiporder = tre.get_tip_labels()
        itertree = open(trees, 'r')
    elif isinstance(trees, toytree.ToyTree):
        itertree = iter([trees])
        ntrees = len(trees)
        ntips = trees.ntips
        tiporder = trees.get_tip_labels()
    elif isinstance(trees, toytree.MultiTree):
        itertree = iter(i for i in trees)
        ntrees = len(trees)
        ntips = trees.treelist[0].ntips
        tiporder = trees.treelist[0].get_tip_labels()
    else:
        raise IOError(f"problem with input: {trees}")

    # array to store results
    tarr = np.zeros((ntips, ntrees))

    # run non-parallel calculations
    if njobs == 1:
        for tidx, tree in enumerate(itertree):
            data = function(tree, **kwargs)
            tarr[:, tidx] = data[tiporder]

    # or, distribute jobs in parallel (py3 only)
    else:
        pool = ProcessPoolExecutor(njobs)
        rasyncs = {}

        # submit as many jobs as there are cores
        for job in range(njobs):
            tree = next(itertree)
            kwargs.update({"tree": tree})
            rasyncs[job] = pool.submit(function, **kwargs)

        # as each job finished submit a new one until none are left.
        # this allows avoiding loading all trees simultaneously.
        tidx = 0
        while 1:
            # get finished jobs
            finished = [i for i in rasyncs if rasyncs[i].done()]

            # store results and append new job to the engine
            for job in finished:

                # store result
                result = rasyncs[job].result()
                tarr[:, tidx] = result[tiporder]
                tidx += 1
                del rasyncs[job]

                # append new job unless no jobs left
                try:
                    tree = next(itertree)
                    kwargs.update({"tree": tree})
                    rasyncs[job] = pool.submit(function, **kwargs)
                except StopIteration:
                    pass

            # wait until all jobs finish
            if not rasyncs:
                break
        pool.shutdown()

    # close the file handle, if exists
    if hasattr(itertree, "close"):
        itertree.close()

    # calculate summary of DR across the distribution of trees.
    arr = pd.DataFrame(tarr, index=tiporder)
    data = pd.DataFrame({
        'mean': arr.mean(1),
        # 'harmMean': hmean(arr, axis=1),
        'median': arr.median(1),
        'std': arr.std(1),
        '2.5%': np.percentile(arr, 0.025, axis=1),
        '97.5%': np.percentile(arr, 0.975, axis=1),
        'min': arr.min(1),
        'max': arr.max(1),
    })
    return data


if __name__ == "__main__":

    pass



#!/usr/bin/python

"""toytree subpackage for modifying tree topology or features.

>>> tree = toytree.tree.rtree(10)

>>> tree = tree.mod.root()
>>> tree = tree.mod.unroot()
>>> tree = tree.mod.edges_scale_to_root_height()
>>> tree = tree.mod.edges_slide_nodes()
>>> tree = tree.mod.edges_align_tips_by_extending()
>>> tree = tree.mod.edges_align_tips_by_penalized_likelihood()
>>> tree = tree.mod.add_internal_node()
>>> tree = tree.mod.add_leaf_node()
>>> tree = tree.mod.move_clade()
>>> tree = tree.mod._iter_spr_unrooted()
>>> tree = tree.mod.move_spr_unrooted()

# raised to ToyTree level
>>> tree.root()
>>> tree.unroot()
"""

from toytree.mod._src.mod_edges import *
from toytree.mod._src.mod_topo import *
from toytree.mod._src.root_unroot import *
from toytree.mod._src.root_funcs import *
# from toytree.mod._src.tree_move import *
#     move_spr,
#     move_spr_iter,
#     move_nni,
#     move_nni_iter,
# )



#!/usr/bin/env python

"""Dispatched renderer for toytree marks following the style of toyplot

TODO
----
  - fixed-order extension to tip positions for missing labels..?
  - container tree to Mark
  - Circle tree does not project y-axis...


<g class='pie-0'b style='stroke: black; stroke-width: 1'>
    <path d="M 1 0 A 1 1 0 0 1 0.8 0.5 L 0 0" style='fill:red'><path>
    <path d="M 1 0 A 0 1 0 0 1 0.8 0.5 L 0 0" style='fill:pink'><path>
                     r r x l s  x   y
</g>           |    |                  |
               move to start position on arc.
                    |                  |
                    arc: rx ry x-axis-rotation large-arc-floag sweep-flag x y
                         radius    always 0        %>50 or no    always 1  end
                                       |
                                       move to center of circle. 
"""

from typing import List, Dict, Tuple
import functools
import xml.etree.ElementTree as xml
from multipledispatch import dispatch
from loguru import logger
import numpy as np
import toyplot
from toytree.drawing import ToyTreeMark
from toytree.drawing.src.render_text import render_text
from toytree.drawing.src.render_marker import render_marker
from toytree.color.src.concat import concat_style_fix_color
from toytree.layout.src.get_edge_midpoints import get_edge_midpoints

# SVG path formats for creating edges in tree drawings
PATH_FORMAT = {
    'c': "M {px:.1f} {py:.1f} L {cx:.1f} {cy:.1f}",  # angular /\ format, good for any layout
    'b1': "M {px:.1f} {py:.1f} C {px:.1f} {cy:.1f}, {px:.1f} {cy:.1f}, {cx:.1f} {cy:.3f}",  # bezier 1 format
    'b2': "M {px:.1f} {py:.1f} C {cx:.1f} {py:.1f}, {cx:.1f} {py:.1f}, {cx:.1f} {cy:.3f}",  # bezier 2 format
    'p1': "M {px:.1f} {py:.1f} L {px:.1f} {cy:.1f} L {cx:.1f} {cy:.1f}",  # |_| phylo/angular format 1
    'p2': "M {px:.1f} {py:.1f} L {cx:.1f} {py:.1f} L {cx:.1f} {cy:.1f}",  # |_| phylo/angular format 2
    'pc': "M {cx:.1f} {cy:.1f} L {dx:.1f} {dy:.1f} A {rr:.1f} {rr:.1f} 0 0 {sweep} {px:.1f} {py:.1f}",  # phylo for circular
}

logger = logger.bind(name="toytree")

# ---------------------------------------------------------------------
# Register multipledispatch to use the toyplot.html namespace
dispatch = functools.partial(dispatch, namespace=toyplot.html._namespace)


# register a _render function for ToyTreeMark objects
@dispatch(toyplot.coordinates.Cartesian, ToyTreeMark, toyplot.html.RenderContext)
def _render(axes, mark, context):
    RenderToytree(axes, mark, context)
# ---------------------------------------------------------------------


class RenderToytree:
    """Class with functions to add ToyTree class to the HTML DOM.

    Organized class to call within _render. The top level canvas
    element is .context. From this parent xml subelements are
    added to build the drawing. The toytree mark is in .mark.
    """
    def __init__(self, axes, mark, context):
        # existing
        self.mark = mark
        self.axes = axes
        self.context = context

        # start rendering of this Mark by connecting to context (Canvas)
        self.mark_xml: xml.SubElement = xml.SubElement(
            context.parent, "g",
            id=context.get_id(self.mark),
            attrib={"class": "toytree-mark-Toytree"},
        )

        # to be constructed (are these reused?)
        self.edges_xml: xml.SubElement = None
        self.nodes_xml: xml.SubElement = None
        self.admix_xml: xml.SubElement = None
        self.align_xml: xml.SubElement = None

        # data...
        self.nodes_x: np.ndarray = None
        self.nodes_y: np.ndarray = None
        self.tips_x: np.ndarray = None
        self.tips_y: np.ndarray = None
        self.radii: np.ndarray = None

        # construction funcs
        self.project_coordinates()
        self.build_dom()

    def project_coordinates(self):
        """Store node coordinates (data units) projected to px units."""
        self.nodes_x = self.axes.project('x', self.mark.ntable[:, 0])
        self.nodes_y = self.axes.project('y', self.mark.ntable[:, 1])
        self.tips_x = self.axes.project('x', self.mark.ttable[:, 0])
        self.tips_y = self.axes.project('y', self.mark.ttable[:, 1])
        mcoords = get_edge_midpoints(
            self.mark.etable, self.mark.ntable, self.mark.layout, self.mark.edge_type)
        self.mid_x = self.axes.project('x', mcoords[:, 0])
        self.mid_y = self.axes.project('y', mcoords[:, 1])

    def build_dom(self):
        """Creates DOM of xml.SubElements in self.context."""
        self.mark_edges()
        self.mark_align_edges()
        self.mark_admixture_edges()
        self.mark_nodes()
        self.mark_node_labels()

        # for multitrees tips are sometimes not drawn.
        self.mark_tip_labels()

    def get_paths(self) -> Tuple[List[str], List[str]]:
        """Return paths and keys in idx order.

        This will build the d="..." path string for the SVG lines
        for edges of the tree. Depending on the edge_type this can be
        relatively simple or more complex.
        """
        # modify order of x or y shift of edges for p,b types.
        if self.mark.edge_type in ('p', 'b'):

            # phylo |_| edge type for cirular trees:
            if self.mark.layout[0] == 'c':
                path_format = PATH_FORMAT["pc"]

                # get differences of nodes relative to root. Note: this is
                # b/c origin is not at (0, 0) if [x,y]baseline was used.
                xdiffs = self.mark.ntable[:, 0] - self.mark.ntable[-1, 0]
                ydiffs = self.mark.ntable[:, 1] - self.mark.ntable[-1, 1]
                radii = np.sqrt(xdiffs ** 2 + ydiffs ** 2)

                # get angles of nodes relative to root in radians (0, 2pi)
                radians = np.arctan2(ydiffs, xdiffs)
                radians[radians < 0] = (2 * np.pi) + radians[radians < 0]

            # phylo or bezier type for non-circular trees. Select the
            # appropriate type given orientation: p1, p2, or b1, b2
            elif self.mark.layout in ('u', 'd'):
                path_format = PATH_FORMAT[f"{self.mark.edge_type}2"]
            else:
                path_format = PATH_FORMAT[f"{self.mark.edge_type}1"]

        # cladogram (\/) style, simplest for any layout
        else:
            path_format = PATH_FORMAT[self.mark.edge_type]

        # store paths here
        paths = []
        keys = []

        # TODO: change here if you want to show the root edge...
        for idx in range(self.mark.nnodes - 1):
            cidx, pidx = self.mark.etable[idx]
            child_x, child_y = self.nodes_x[cidx], self.nodes_y[cidx]
            parent_x, parent_y = self.nodes_x[pidx], self.nodes_y[pidx]

            # for simple edge format (e.g., 'c', 'b')
            if "A" not in path_format:
                keys.append(f"{pidx},{cidx}")
                paths.append(
                    path_format.format(**{
                        'cx': child_x, 'cy': child_y,
                        'px': parent_x, 'py': parent_y,
                    })
                )

            # for 'p' edge format
            else:
                # get radius at parent's level
                xdiff = parent_x - self.nodes_x[-1]
                ydiff = parent_y - self.nodes_y[-1]
                parent_radius = np.sqrt(xdiff ** 2 + ydiff ** 2)

                mid_x = radii[pidx] * np.cos(radians[cidx])
                mid_y = radii[pidx] * np.sin(radians[cidx])

                # project mid points into coordinate space
                px_mid_x = self.axes.project('x', mid_x)
                px_mid_y = self.axes.project('y', mid_y)

                # logger.warning(f"{cidx} {radians[pidx]:.2f} {radians[cidx]:.2f} sweep={int(radians[pidx] < radians[cidx])}")
                keys.append(f"{pidx},{cidx}")
                paths.append(
                    path_format.format(**{
                        'cx': child_x,
                        'cy': child_y,
                        'px': parent_x,
                        'py': parent_y,
                        'dx': px_mid_x,
                        'dy': px_mid_y,
                        'rr': parent_radius,
                        'sweep': int(radians[pidx] < radians[cidx]),  # 1=counter-clockwise, 0=clockwise
                    })
                )
            # keys.append('root-edge')
            # paths.append(
            #     path_format.format(**{
            #         'cx': self.nodes_x[-1],
            #         'cy': self.nodes_y[-1],
            #         'px': self.nodes_x[-1],
            #         'py': self.nodes_y[-1] + self.mark.root_dist,
            #     })
            # )
        return paths, keys

    def mark_edges(self) -> None:
        """Create SVG paths for each tree edge as class toytree-Edges"""
        # get paths based on edge type and layout
        paths, keys = self.get_paths()

        # !always pop 'fill' to set it to 'fill:none' below (no fill btwn edges).
        _ = self.mark.edge_style.pop('fill', None)

        # render the edge group.
        # TODO: consider adding stroke-linejoin:round
        self.edges_xml = xml.SubElement(
            self.mark_xml, "g",
            attrib={"class": "toytree-Edges"},
            style=concat_style_fix_color(self.mark.edge_style, "fill:none")
        )

        # get shared versus unique styles (EdgeStyles)
        unique_styles = get_unique_edge_styles(self.mark)
        # unique_styles.append({"stroke-width": 1.0})
        # logger.info(unique_styles)

        # render the edge paths
        for idx, path in enumerate(paths):
            xml.SubElement(
                self.edges_xml, "path",
                d=path,
                id=keys[idx],
                style=concat_style_fix_color(unique_styles[idx])
            )

    def mark_nodes(self) -> None:
        """Create marker elements for each node in class toytree-Nodes.

        Stores ids to the nodes which in theory can allow for
        downstream JS interactivity.
        """
        # Group all Nodes with shared style applied
        self.nodes_xml = xml.SubElement(
            self.mark_xml, "g",
            attrib={"class": "toytree-Nodes"},
            style=concat_style_fix_color(self.mark.node_style)
        )

        # skip drawing any nodes if node_mask=True or all node_sizes=0
        if (self.mark.node_sizes == 0).all():
            return
        if not self.mark.node_mask.any():
            return

        # get dict with only styles not shared by all Nodes. This sets
        # fill-opacity to False (special) if there is a shared style
        # for node_style.fill-opacity (except for fill=transparent)
        unique_styles = get_unique_node_styles(self.mark)

        # get nnodes to draw (fewer if drawing edges or unrooted)
        if self.mark.node_as_edge_data:
            nedges = self.mark.nnodes - 2
            if np.count_nonzero(self.mark.etable[:, 1] == self.mark.etable[-1, 1]) > 2:
                nedges += 1
            nmarkers = nedges
            xcoords = self.mid_x
            ycoords = self.mid_y
        else:
            nmarkers = self.mark.nnodes
            xcoords = self.nodes_x
            ycoords = self.nodes_y

        # add node markers in reverse idx order (levelorder traversal)
        # for nidx in range(self.mark.nnodes):
        for nidx in range(nmarkers):

            # skip if node is masked
            if not self.mark.node_mask[nidx]:
                continue

            # create marker with shape and size, e.g., <marker='o' size=12>
            marker = toyplot.marker.create(
                shape=self.mark.node_markers[nidx],
                size=self.mark.node_sizes[nidx],
            )

            # create the marker
            marker_xml = xml.SubElement(
                self.nodes_xml, "g",
                attrib={"id": f"Node-{nidx}"},
                style=concat_style_fix_color(unique_styles[nidx]),
            )

            # if style string is empty then remove it.
            if not marker_xml.attrib['style']:
                _ = marker_xml.attrib.pop('style')

            # optionally add a title (hover) here only if there is no
            # node_label, otherwise we put the hover on the label later.
            if self.mark.node_hover is not None:
                if self.mark.node_labels is None:
                    xml.SubElement(marker_xml, "title").text = (
                        self.mark.node_hover[nidx])

            # project marker in coordinate space
            transform = "translate({:.6g},{:.6g})".format(
                xcoords[nidx],
                ycoords[nidx],
            )
            if marker.angle:
                transform += " rotate({:.3f})".format(-marker.angle)
            marker_xml.set("transform", transform)

            # get shape type
            render_marker(marker_xml, marker)

    def mark_node_labels(self) -> None:
        """Create Node labels in toytree-NodeLabels using render_text"""
        if self.mark.node_labels is None:
            return

        # shared styles popped from text boxes AFTER positioning
        shared_style = {
            "font-size": self.mark.node_labels_style['font-size'],
            "font-weight": self.mark.node_labels_style['font-weight'],
            "font-family": self.mark.node_labels_style['font-family'],
            "vertical-align": "baseline",
            "white-space": "pre",
        }

        # create NodeLabels group element
        node_labels_xml = xml.SubElement(
            self.mark_xml, "g",
            attrib={"class": "toytree-NodeLabels"},
            style=concat_style_fix_color(shared_style, "stroke:none"),
        )

        # get nnodes to draw (fewer if drawing edges or unrooted)
        if self.mark.node_as_edge_data:
            nedges = self.mark.nnodes - 2
            if np.count_nonzero(self.mark.etable[:, 1] == self.mark.etable[-1, 1]) > 2:
                nedges += 1
            nmarkers = nedges
            xcoords = self.mid_x
            ycoords = self.mid_y
        else:
            nmarkers = self.mark.nnodes
            xcoords = self.nodes_x
            ycoords = self.nodes_y

        # apply unique styles to each node label
        for idx in range(nmarkers):
            # masked label
            if not self.mark.node_mask[idx]:
                continue

            # get the label
            label = self.mark.node_labels[idx]
            title = None if self.mark.node_hover is None else self.mark.node_hover[idx]

            # always align node labels in middle of marker (?)
            default = {
                "-toyplot-vertical-align": "middle", # baseline not supported.
                "text-anchor": "middle", # start is another option.
            }
            nlstyle = toyplot.style.combine(self.mark.node_labels_style, default)
            render_text(
                root=node_labels_xml,
                text=label,
                xpos=xcoords[idx],
                ypos=ycoords[idx],
                angle=0,
                attributes={"class": "toytree-NodeLabel"},
                style=nlstyle,
                title=title,
            )

    def mark_tip_labels(self):
        """Create tip labels in toytree-TipLabels using render_text."""
        if self.mark.tip_labels is None:
            return

        # if angles are None then compute them now given the layout.

        # TipLabels style keys are popped from TipLabel styles, but
        # do not include positioning styles, which must be passed on.
        # shared_style = self.mark.tip_labels_style.copy()

        # base styles for the <g>, this includes some shared styles 
        # such as: fill, stroke, but cannot do baseline-shift, 
        # text-anchor, or anchor-shift (positioning text styles)
        shared_style = {
            "font-size": self.mark.tip_labels_style['font-size'],
            "font-weight": self.mark.tip_labels_style['font-weight'],
            "font-family": self.mark.tip_labels_style['font-family'],
            "fill": self.mark.tip_labels_style['fill'],
            "vertical-align": "baseline",
            "white-space": "pre",
        }

        # Make the <g> TipLabels group element
        tips_xml = xml.SubElement(
            self.mark_xml, "g",
            attrib={"class": "toytree-TipLabels"},
            style=concat_style_fix_color(shared_style, "stroke:none"),
        )

        # add <text> tip markers from 0 to ntips
        for idx, tip in enumerate(self.mark.tip_labels):

            # get coordinates of tips
            cxx = self.nodes_x[idx]
            cyy = self.nodes_y[idx]
            if self.mark.tip_labels_align:
                cxx = self.tips_x[idx]
                cyy = self.tips_y[idx]

            # style dict for this specific <text> string. This needs to
            # ALSO include the font-size, font-weight, etc., so that we
            # can calculate ... but then it will be popped from CSS.
            tstyle = self.mark.tip_labels_style.copy()
            if self.mark.tip_labels_colors is not None:
                _ = tstyle.pop("fill", None)
                tstyle["fill"] = self.mark.tip_labels_colors[idx]

            # assign tip layout positioning
            offset = toyplot.units.convert(
                tstyle["-toyplot-anchor-shift"], "px", "px",
            )

            if self.mark.layout in 'rd':
                angle = self.mark.tip_labels_angles[idx]

            # reverse labels
            elif self.mark.layout in 'lu':
                angle = self.mark.tip_labels_angles[idx]
                tstyle['-toyplot-anchor-shift'] = -offset
                tstyle['text-anchor'] = "end"

            # 'c' or unrooted
            else:
                angle = self.mark.tip_labels_angles[idx]
                if 90 < angle < 270:
                    tstyle['text-anchor'] = "end"
                    tstyle['-toyplot-anchor-shift'] = -offset
                    angle -= 180

            # add text
            render_text(
                root=tips_xml,
                text=tip,
                xpos=cxx,
                ypos=cyy,
                angle=angle,
                attributes={"class": "toytree-TipLabel"},
                style=tstyle,
            )

    def mark_align_edges(self):
        """Create SVG paths for each tip to 0 or radius.

        Currently only group styles are suppored for aligned edges.
        """
        # get paths based on edge type and layout
        if self.mark.tip_labels_align:
            apaths = []
            for tidx, _ in enumerate(self.mark.tip_labels_angles):

                adict = {
                    'cx': self.nodes_x[tidx],
                    'cy': self.nodes_y[tidx],
                    'px': self.tips_x[tidx],
                    'py': self.tips_y[tidx],
                }
                path = PATH_FORMAT['c'].format(**adict)
                apaths.append(path)

            # render the edge group
            self.align_xml = xml.SubElement(
                self.mark_xml, "g",
                attrib={"class": "toytree-AlignEdges"},
                style=concat_style_fix_color(self.mark.edge_align_style),
            )

            # render the edge paths
            for _, path in enumerate(apaths):
                xml.SubElement(self.align_xml, "path", d=path)

    def mark_admixture_edges(self):
        """Create SVG paths for admixture edges.

        The edge takes the same style as the edge_type of the tree.

        (SRC, DEST, PROP, {STYLE}, TEXT)
        """
        if self.mark.admixture_edges is None:
            return

        # iterate over colors for subsequent edges unless provided
        default_admix_edge_style = {
            "stroke": 'rgb(90.6%,54.1%,76.5%)',
            "stroke-width": 5,
            "stroke-opacity": 0.6,
            "stroke-linecap": "round",
            "fill": "none",
            "font-size": "14px"
        }

        # create edge group element
        self.admix_xml = xml.SubElement(
            self.mark_xml, 'g',
            attrib={'class': 'toytree-AdmixEdges'},
            style=concat_style_fix_color(default_admix_edge_style)
        )

        # get position of 15% tipward from source point
        path_format = [
            "M {sdx:.2f} {sdy:.2f}",
            "L {sux:.2f} {suy:.2f}",
            "L {ddx:.2f} {ddy:.2f}",
            "L {dux:.2f} {duy:.2f}",
        ]

        # ensure admixture_edges is a list of tuples
        if not isinstance(self.mark.admixture_edges, list):
            self.mark.admixture_edges = [self.mark.admixture_edges]

        # drwa each edge
        for aedge in self.mark.admixture_edges:

            # check if nodes have an overlapping interval
            src, dest, *details = aedge

            # parse optional prop float, edgestyle dict, and label text
            aprop = 0.5
            estyle = {}
            label = ""
            if len(details) == 1:
                aprop = details[0]
            elif len(details) == 2:
                aprop, estyle = details
            elif len(details) == 3:
                aprop, estyle, label = details

            # get their parents coord positions
            try:
                psrc = self.mark.etable[self.mark.etable[:, 0] == src, 1][0]
            except IndexError:
                psrc = self.mark.nnodes - 1
            try:
                pdest = self.mark.etable[self.mark.etable[:, 0] == dest, 1][0]
            except IndexError:
                pdest = self.mark.nnodes - 1

            # if only one midpoint then expand to use same for both edges.
            shared = False
            if isinstance(aprop, (int, float)):
                shared = True
                aprop = (aprop, aprop)

            # separate for each layout b/c its haaaard.
            if self.mark.layout in ("r", "l"):
                src_x, src_y = self.nodes_y[src], self.nodes_x[src]
                dst_x, dst_y = self.nodes_y[dest], self.nodes_x[dest]
                p_src_x, p_src_y = self.nodes_y[psrc], self.nodes_x[psrc]
                p_dst_x, p_dst_y = self.nodes_y[pdest], self.nodes_x[pdest]

                # ...
                if self.mark.layout == 'r':
                    disjoint = (p_src_y >= dst_y) or (src_y <= p_dst_y)
                    sign = 1
                else:
                    disjoint = (p_src_y >= dst_y) or (src_y <= p_dst_y)
                    sign = -1

                # ...
                if disjoint or (not shared):
                    src_mid_y = src_y - sign * (abs(src_y - p_src_y) * aprop[0])
                    dest_mid_y = dst_y - sign * (abs(dst_y - p_dst_y) * aprop[1])
                else:
                    # get height of the admix line at midshared.
                    amin = min([src_y, dst_y])
                    amax = max([p_src_y, p_dst_y])
                    admix_ymid = amin + (amax - amin) * aprop[0]
                    dest_mid_y = src_mid_y = admix_ymid

            elif self.mark.layout in ("u", "d"):
                # get x and y of source and destination nodes
                src_x, src_y = self.nodes_x[src], self.nodes_y[src]
                dst_x, dst_y = self.nodes_x[dest], self.nodes_y[dest]

                # get x and y of PARENTS of source and destination nodes
                p_src_x, p_src_y = self.nodes_x[psrc], self.nodes_y[psrc]
                p_dst_x, p_dst_y = self.nodes_x[pdest], self.nodes_y[pdest]

                # check whether the edges overlap, in which case we will
                # draw a straight line between them, otherwise the line
                # will be angled. Straight it preferred.
                if self.mark.layout == "d":
                    disjoint = (dst_y <= p_src_y) or (src_y <= p_dst_y)
                    sign = 1
                else:
                    disjoint = (dst_y >= p_src_y) or (src_y >= p_dst_y)
                    sign = -1

                if disjoint or (not shared):
                    src_mid_y = src_y - sign * (abs(src_y - p_src_y) * aprop[0])
                    dest_mid_y = dst_y - sign * (abs(dst_y - p_dst_y) * aprop[1])
                else:
                    amin = min([src_y, dst_y])
                    amax = max([p_src_y, p_dst_y])
                    admix_ymid = amin - sign * abs(amax - amin) * aprop[0]
                    dest_mid_y = src_mid_y = admix_ymid

            # project angle of up/down lines towards parent nodes.
            if self.mark.edge_type == "c":

                # angle from src to src parent
                if (p_src_x - src_x) == 0:
                    x_shift_src_mid = 0
                else:
                    theta = np.arctan((p_src_y - src_y) / (p_src_x - src_x))
                    x_shift_src_mid = (src_mid_y - src_y) / np.tan(theta)

                # angle from dest to dest parent
                if (p_dst_x - dst_x) == 0:
                    x_shift_dest_mid = 0
                else:
                    theta = np.arctan((p_dst_y - dst_y) / (p_dst_x - dst_x))
                    x_shift_dest_mid = (dest_mid_y - dst_y) / np.tan(theta)
                xend = p_dst_x

            else:
                x_shift_dest_mid = 0
                x_shift_src_mid = 0
                xend = dst_x

            # build the SVG path
            if self.mark.layout in ("r", "l"):
                edge_dict = {
                    'sdy': src_x,  # + x_shift_src_tip + snudge,
                    'sdx': src_y,  # src_tip_y,
                    'suy': src_x + x_shift_src_mid,
                    'sux': src_mid_y,  # admix_ymid,
                    'ddy': dst_x + x_shift_dest_mid,
                    'ddx': dest_mid_y,  # admix_ymid,
                    'duy': xend,
                    'dux': p_dst_y,  # dest_tip_y,
                }
                # tri_dict = {
                #     'x0': admix_ymid - 6,
                #     'x1': admix_ymid + 6,
                #     'x2': admix_ymid,
                #     'y0': np.mean([edge_dict['suy'], edge_dict['ddy']]) - 6,
                #     'y1': np.mean([edge_dict['suy'], edge_dict['ddy']]) - 6,
                #     'y2': np.mean([edge_dict['suy'], edge_dict['ddy']]) + 8,
                # }

            else:
                edge_dict = {
                    'sdx': src_x,  # + x_shift_src_tip + snudge,
                    'sdy': src_y,  # src_tip_y,

                    'sux': src_x + x_shift_src_mid,
                    'suy': src_mid_y,  # admix_ymid,

                    'ddx': dst_x + x_shift_dest_mid,
                    'ddy': dest_mid_y,  # admix_ymid,

                    'dux': xend,
                    'duy': p_dst_y,  # dest_tip_y,
                }

                # TODO: not finished aligning triangle/arrow
                # tri_dict = {
                #     'y0': admix_ymid - 6,
                #     'y1': admix_ymid + 6,
                #     'y2': admix_ymid,
                #     'x0': np.mean([edge_dict['suy'], edge_dict['ddy']]) - 6,
                #     'x1': np.mean([edge_dict['suy'], edge_dict['ddy']]) - 6,
                #     'x2': np.mean([edge_dict['suy'], edge_dict['ddy']]) + 8,
                # }

            # EDGE path
            path = " ".join(path_format).format(**edge_dict)

            # TODO: split style not needed.
            # estyle['stroke'] = split_rgba_style(estyle['stroke'])
            xml.SubElement(
                self.admix_xml, "path",
                d=path,
                style=concat_style_fix_color(estyle),
            )

            lstyle = estyle.copy()
            # LABEL
            if label is not None:

                # RENDER edge label
                lstyle['fill'] = '#262626'
                lstyle['fill-opacity'] = '1.0'
                lstyle['stroke'] = "none"
                lstyle['text-anchor'] = 'middle'

                # position
                if self.mark.layout in ("r", "l"):
                    xtext = np.mean([src_x + x_shift_src_mid, dst_x + x_shift_dest_mid])
                    ytext = np.mean([src_mid_y, dest_mid_y])
                    xtext += 12
                else:
                    ytext = np.mean([src_x + x_shift_src_mid, dst_x + x_shift_dest_mid])
                    xtext = np.mean([src_mid_y, dest_mid_y])

                xml.SubElement(
                    self.admix_xml,
                    "text",
                    x=f"{ytext:.2f}",
                    y=f"{xtext:.2f}",
                    style=concat_style_fix_color(lstyle),
                ).text = str(label)


# HELPER FUNCTIONS ----------------------
def get_unique_edge_styles(mark) -> List[Dict]:
    """Return dicts of unique edge stroke or width or each Node

    Reduces node styles to prevent redundancy in HTML.
    """
    # minimum styling of node markers
    unique_styles = [{} for i in range(mark.etable.shape[0])]

    # if edge widths and edge colors are both None then just return
    if (mark.edge_colors is None) & (mark.edge_widths is None):
        return unique_styles

    # iterate through styled node marks to get shared styles and expand axes
    for idx in range(mark.etable.shape[0]):

        if mark.edge_widths is not None:
            unique_styles[idx]['stroke-width'] = mark.edge_widths[idx]

        if mark.edge_colors is not None:
            unique_styles[idx]['stroke'] = mark.edge_colors[idx]

        # concat_style_to_str2() will interpret False to write no opacity style
        if mark.edge_style.get("stroke-opacity") is not None:
            unique_styles[idx]['stroke-opacity'] = False

    return unique_styles


def get_unique_node_styles(mark) -> List[Dict]:
    """Return dicts of unique styles (colors) for each Node."""
    unique_styles = [{} for i in range(mark.nnodes)]

    # only if variable tho
    if mark.node_colors is None:
        return unique_styles

    # get fill which will be split later into rgb, a
    for idx in range(mark.nnodes):
        unique_styles[idx]['fill'] = mark.node_colors[idx]
        if mark.node_style.get("fill-opacity") is not None:
            # special handling for transparency...
            unique_styles[idx]['fill-opacity'] = False
    return unique_styles


def test1():
    tree = toytree.rtree.rtree(8, seed=123)
    kwargs = dict(
        ts='p',
        layout='r', #'c0-95',
        tip_labels_align=True,
        edge_type='p',
        width=400,
        height=400,
        node_sizes=18,
        node_colors=None,
        node_mask=False,
        node_hover=True,
        node_labels=True,
        node_labels_style={"font-size": 15, "fill": "black"},
        edge_style={"stroke-width": 3.5},
        node_style={"stroke-width": 2},
        tip_labels_style={"font-size": 15, "-toyplot-anchor-shift": 15},
        # tip_labels_colors=[toytree.color.COLORS1[i] for i in range(tree.ntips)],
        # edge_colors=['red'] + [toytree.color.COLORS1[0]] + ['black'] * (tree.nnodes - 2),
    )
    tree._draw_browser(**kwargs)


def test2():
    # generate a random species tree with 10 tips and a crown age of 10M generations
    tree = toytree.rtree.unittree(10, treeheight=1e6, seed=123)
    # create a new tree copy with Ne values mapped to nodes
    vtree = tree.set_node_data(
        feature="Ne",
        data={i: 2e5 for i in (6, 7, 8, 9, 12, 15, 17)},
        default=1e4,
    )
    vtree._draw_browser(ts='p', )#admixture_edges=[(0, 12, 0.5, {}, "word")]);


if __name__ == "__main__":

    import toytree
    toytree.set_log_level("DEBUG")
    test2()



#!/usr/bin/env python

"""...

"""

from typing import Tuple, List, Mapping, Any
import numpy as np
import toyplot
from toyplot.mark import Mark
from loguru import logger

logger = logger.bind(name="toytree")


class AnnotationMarker(Mark):
    """TipMarkers are Point markers with a custom render function that
    allows for adding a transform to shift position by px units.
    """
    def __init__(
        self,
        ntable: np.ndarray,
        sizes: np.ndarray,
        shapes: np.ndarray,
        colors: np.ndarray,
        opacity: float,
        xshift: float,
        yshift: float,
        style: Mapping[str, Any],
    ):
        Mark.__init__(self, annotation=True)
        self._coordinate_axes = ['x', 'y']
        self.ntable = ntable
        self.sizes = sizes
        self.shapes = shapes
        self.colors = colors
        self.opacity = opacity
        self.xshift = xshift
        self.yshift = yshift
        self.style = style

    def domain(self, axis: str) -> Tuple[float, float]:
        """The markers define the domain of data"""
        index = self._coordinate_axes.index(axis)
        domain = toyplot.data.minimax(self.ntable[:, index])
        return domain

    def extents(self, axis: str) -> Tuple[Tuple[np.ndarray], Tuple[np.ndarray]]:
        """The marker shape and size define the extents of the data."""
        coords = ()
        if 'x' in axis:
            coords += (self.ntable[:, 0], )
        if 'y' in axis:
            coords += (self.ntable[:, 1], )

        # get marker/label extents and update minmax for each feature
        extents = [np.zeros(self.ntable.shape[0])] * 4
        extents = set_marker_extents(self, extents)
        return coords, extents


class AnnotationRect(Mark):
    """Rectangles annotation Mark.

    The rectangle annotation is used to create clade outlines, bars.
    These Marks can be defined both in terms of data width,height as
    well as in terms of additional canvas px units.
    """
    def __init__(
        self,
        ntable: np.ndarray,
        xtable: np.ndarray,
        ytable: np.ndarray,
        # heights: np.ndarray,
        colors: np.ndarray,
        opacity: float,
        xshift: float,
        yshift: float,
        style: Mapping[str, Any],
    ):
        Mark.__init__(self, annotation=True)
        self._coordinate_axes = ['x', 'y']
        self.ntable = ntable
        """: array of (nmarks, 2) ..."""
        self.xtable = xtable
        self.ytable = ytable

        self.colors = colors
        self.opacity = opacity
        self.xshift = xshift
        self.yshift = yshift
        self.style = style

    def domain(self, axis: str) -> Tuple[float, float]:
        """The markers define the domain of data"""
        index = self._coordinate_axes.index(axis)
        if index == 0:
            return toyplot.data.minimax(self.xtable)
        return toyplot.data.minimax(self.ytable)
        # domain = toyplot.data.minimax([
        #     self.ntable[:, index], self.bar_min, self.bar_max,
        #     self.ntable[:, index] + self.bar_max,
        # ])
        # # print(f"X--- {index}, {self.widths}, {self.heights}\n\n")
        # return domain


def set_marker_extents(mark: Mark, extents: List[np.ndarray]) -> List[np.ndarray]:
    """Return extents of node label text string extents.
    """
    # markers are symmetrical NxN unless 'rNxM'
    yext = np.zeros(mark.ntable.shape[0])
    xext = np.zeros(mark.ntable.shape[0])

    iter_data = zip(mark.shapes, mark.sizes)
    for idx, (marker, size) in enumerate(iter_data):
        if marker.startswith("r"):
            width, height = [int(i) for i in marker[1:].split("x")]
        else:
            width, height = 1, 1
        yext[idx] = height * size
        xext[idx] = width * size

    # extent is half the markers dimension in either direction
    width = mark.style.get("stroke-width", 0)
    xext = xext / 2. + width
    yext = yext / 2. + width

    # set extents (implement shift)
    extents[0] = np.min([extents[0], -xext + mark.xshift], axis=0)
    extents[1] = np.max([extents[1], xext + mark.xshift], axis=0)
    extents[2] = np.min([extents[2], -yext + mark.yshift], axis=0)
    extents[3] = np.max([extents[3], yext + mark.yshift], axis=0)
    return extents


if __name__ == "__main__":
    pass



#!/usr/bin/env python

"""Testing simple alternative validation

"""

from typing import Union, Sequence, Mapping, Any, Dict, TypeVar, Tuple, Optional, List
import numpy as np
import toyplot
from loguru import logger

from toytree.color import ToyColor
from toytree.utils import ToytreeError, ToyColorError
from toytree.style.src.validate_utils import check_arr
from toytree.style.src.style_base import NodeStyle, NodeLabelStyle, TreeStyle
from toytree.style import get_range_mapped_feature, get_color_mapped_feature

ToyTree = TypeVar("ToyTree")
Color = TypeVar("Color", str, tuple, np.ndarray)
logger = logger.bind(name="toytree")

__all__ = [
    "validate_mask",
    "validate_numeric",
    "validate_markers",
    "validate_colors",
    "validate_hover",
    "validate_labels",
    # "validate_style",
    # "validate_node_markers",
    # "validate_node_hover",
    # "validate_node_style",
]


def validate_mask(
    tree: ToyTree,
    tree_style: Optional[TreeStyle] = None,
    style: Optional[Mapping[str, Any]] = None
) -> np.ndarray:
    """Sets node_mask to ndarray[bool] size=nnodes.

    Mask is a bool array in idx order where False hides node markers.

    Supported args
    --------------
    None: Use default
    True: mask all nodes (return False array)
    False: show all nodes (return True array)
    Sequence: custom boolean mask
    Tuple: (bool, bool, bool) for show (tips, internal, root).
    """
    # get user value or, if None, use style base value
    mask = style.get("node_mask")
    if mask is None:
        if isinstance(tree_style, TreeStyle):
            mask = getattr(tree_style, "node_mask")

    # default None masks tip Nodes and shows internal + root
    if mask is None:
        mask = (0, 1, 1)

    # mask=True means mask all Nodes, so return all False.
    if mask is True:
        mask = np.repeat(False, tree.nnodes)

    # mask=False means show all Nodes, so return all True
    elif mask is False:
        mask = np.repeat(True, tree.nnodes)

    # special tuple arg (show_tips, show_internal, show_root)
    elif isinstance(mask, tuple):
        mask = tree.get_node_mask(
            show_tips=mask[0],
            show_internal=mask[1],
            show_root=mask[2],
        )

    # else custom. Check size and type and return as an array.
    mask = check_arr(mask, "node_mask", tree.nnodes, np.bool_)
    return mask


def validate_numeric(
    tree: ToyTree,
    key: str,
    size: int,
    tree_style: Optional[TreeStyle] = None,
    style: Optional[Mapping[str, Any]] = None,
) -> np.ndarray:
    """Sets node_sizes to ndarray[float].

    Union[float, Sequence[float]],

    None: sizes all set to 0 (at rendering they are not drawn)
    int: sizes all set to same.
    Sequence[int] = sizes in Node idx order.
    tuple[str, Tuple] = sizes mapped to range by feature value
    """
    # get values from user values
    values = style.get(key)

    # or, get values from TreeStyle
    if values is None:
        if isinstance(tree_style, TreeStyle):
            values = getattr(tree_style, key)

    # expand value to a Series of sizes
    if values is None:
        # some return None to allow style args to override, others set
        # to zero so that zero is expanded as the value.
        if key in ["edge_widths", "tip_labels_angles"]:  # 'edge_opacity'
            return None
        values = 0
    if isinstance(values, (int, float)):
        values = np.repeat(values, size)

    # expand str feature name to a tuple of (feature, min, max) so that nan is handled below
    elif isinstance(values, str):
        if size == tree.ntips:
            data = tree.get_tip_data(values)
            values = (values, data.min(), data.max())
        else:
            data = tree.get_node_data(values)
            values = (values, data.min(), data.max())
        # if values == "node_sizes":
        #     values = (values, 5., 20.)
        # elif values == "edge_widths":
        #     values = (values, 2., 5.)
        # else:
        #     values = (values, )

    # expand as value mapped
    if isinstance(values, tuple) and isinstance(values[0], str):  # len(values) <= 4:
        feature, *args = values
        # defaults selected for marker sizes
        min_default = 2 if feature == "edge_widths" else 5
        max_default = 5 if feature == "edge_widths" else 20
        kwargs = dict(zip(("min_value", "max_value", "nan_value"), args))
        kwargs["min_value"] = kwargs.get("min_value", min_default)
        kwargs["max_value"] = kwargs.get("max_value", max_default)
        kwargs["nan_value"] = kwargs.get("nan_value", 0.0)
        kwargs["tips_only"] = True if size == tree.ntips else False
        return get_range_mapped_feature(tree, feature, **kwargs)

    # validate and return as ndarray
    return check_arr(values, key, size, (int, float, np.integer))


def validate_markers(
    tree: ToyTree,
    key: str,
    size: int,
    tree_style: Optional[TreeStyle] = None,
    style: Optional[Mapping[str, Union[int, Sequence[int]]]] = None,
) -> np.ndarray:
    """Sets node_markers to ndarray[str] or ndarray[Marker].

    node_markers: Union[str, Sequence[str], toyplot.marker.Marker],
    """
    # get user value or style base value
    markers = style.get(key)

    # or tree_style
    if markers is None:
        if isinstance(tree_style, TreeStyle):
            markers = getattr(tree_style, key)

    if markers is None:
        markers = "o"
    if isinstance(markers, (str, toyplot.marker.Marker)):
        markers = np.repeat(markers, size)
    markers = check_arr(markers, key, size, (str, toyplot.marker.Marker))
    return markers


def validate_colors(
    tree: ToyTree,
    key: str,
    size: int,
    tree_style: Optional[TreeStyle] = None,
    style: Optional[Mapping[str, Any]] = None,
) -> Tuple[Union[np.ndarray, None], Union[None, Color]]:
    """Set .node_colors to type=ndarray size=nnodes or None.

    If only one color was entered then store as style.fill.
    Only expand .node_colors into an array of size nnodes if there
    is variation among nodes.

    node_colors: Union[None, Color, Sequence[Color], tuple[str, str]],

    Args
    ----
    None: None. Node color will use node_style.fill
    str: set all as ToyColor(str) unless str is a feature name
    ndarray dtype: set all as ToyColor(ndarray)
    tuple w/ len > 2:
    tuple w/ len == 2:
    colormap: not supported. Use (feature, colormap)
    """
    # get node_colors value, else get from TreeStyle, else None.
    colors = style.get(key)

    # if None then try to get from TreeSTyle
    if colors is None:
        if isinstance(style, TreeStyle):
            colors = getattr(style, key, None)
        elif tree_style is not None:
            colors = getattr(tree_style, key, None)

    # if None then Nodes will be colored by style.fill during render
    if colors is None:
        return None, None

    # do colormapping if colors is a str feature name
    if isinstance(colors, str) and colors in tree.features:
        # do not colormap feature if feature is already a valid list of colors
        try:
            colors = ToyColor.color_expander(tree.get_node_data(colors))
        # else colormap it
        except ToyColorError:
            colors = (colors, )

    # special (feature, cmap), but not single color as tuple (0, 0, 1, 1).
    if isinstance(colors, tuple) and isinstance(colors[0], str):
        feat = colors[0]
        kw = dict(cmap=None if len(colors) < 2 else colors[1])
        kw["domain_min"] = None if len(colors) < 3 else colors[2]
        kw["domain_max"] = None if len(colors) < 4 else colors[3]
        kw["nan_value"] = None if len(colors) < 5 else colors[4]
        kw["tips_only"] = size == tree.ntips
        return get_color_mapped_feature(tree, feat, **kw), None

    # get user entered value as ToyColor or List[ndarray]
    # e.g., str, tuple, or ndarray[r,g,b,a] -> ToyColor
    # e.g., List[str], pd.Series[ToyColor], ndarray[void] -> List[ToyColor]
    colors = ToyColor.color_expander(colors)

    # if single color return None. The node_style validator will recognize
    # None and check orig node_color arg to override node_style.fill.
    if isinstance(colors, ToyColor):
        return None, colors

    # convert the List[ToyColor] to an ndarray of dtype np.void
    colors = check_arr(colors, key, size=size, ctype=np.void)
    return colors, None


def validate_hover(
    tree: ToyTree,
    key: str,
    size: int,
    float_format: str = "{:.12g}",
    tree_style: Optional[TreeStyle] = None,
    style: Optional[Mapping[str, Any]] = None,
) -> np.ndarray:
    """Return node_hover as an ndarray of dtype str w/ float formatted vals.

    node_hover: Union[None, bool, str, Sequence[str]],

    None: no hover
    False: no hover
    True: hover with all features
    str: hover w/ just this feature
    Series[str]: hover w/ just these features
    """
    # get user value or style base value
    node_hover = style.get(key)

    # or from treestyle
    if node_hover is None:
        if isinstance(tree_style, TreeStyle):
            node_hover = getattr(tree_style, key)

    # None will not add hover (title) during rendering
    if node_hover is None:
        return None
    if node_hover is False:
        return None

    # True expands to a list of all features on hover.
    if node_hover is True:
        # get all features with default ones listed first
        default_features = ["idx", "dist", "support", "height"]
        non_default = list(set(tree.features) - set(default_features))
        features = default_features + non_default

    # str selects just one feature
    elif isinstance(node_hover, str):
        assert node_hover in tree.features, f"feature {node_hover} not in tree."
        features = [node_hover]

    # Series of features
    else:
        features = []
        for feature in node_hover:
            assert feature in tree.features, f"feature {feature} not in tree."
            features.append(feature)

    # build a multi-line string with feature {key: value} formatted.
    node_hover = {i: [] for i in tree}  # range(size)}
    for node in tree:
        for feature in features:
            val = getattr(node, feature, np.nan)
            if isinstance(val, float):
                val = str(float_format.format(val))
            else:
                val = str(val)
            node_hover[node].append(f"{feature}: {val}")
    # convert to idx ordered List[str]
    node_hover = ["\n".join(node_hover[i]) for i in tree]

    # validate and return as an array
    return check_arr(node_hover, key, size, str)


def validate_labels(
    tree: ToyTree,
    key: str,
    size: int,
    float_format: str = "{:.4g}",
    tree_style: Optional[TreeStyle] = None,
    style: Optional[Mapping[str, Any]] = None,
) -> Union[None, np.ndarray]:
    """Sets labels to np.ndarray[str] or None.

    Also applies floating point string formatting on node_labels. If
    set to None then no labels are added during rendering.

    node_labels: Union[None, bool, str, Sequence[str]],

    Args
    ----
    None: None
    False: None
    True: 'idx' int labels
    str: extract feature
    Series[Any: custom collection of the proper size.
    """
    # get user value or style base value
    labels = style.get(key)

    # or get from treestyle
    if labels is None:
        if isinstance(tree_style, TreeStyle):
            labels = getattr(tree_style, key)

    # Note: don't use 'is in' otherwise error w/ pd.Series type
    if labels is False:
        return None

    # Default None is OK for Node labels, but is "name" for tip labels
    if labels is None:
        return None

    # True: use idxs as labels
    if labels is True:
        if key == "tip_labels":
            labels = "name"
        else:
            labels = "idx"

    # str: extract feature as labels as an np.array
    if isinstance(labels, str):
        labels = tree.get_node_data(labels).values[:size]

    # tuple (feature, str or Callable)
    if isinstance(labels, tuple) and isinstance(labels[0], str):

        # get feature data
        feature, *formatter = labels
        labels = tree.get_node_data(feature).values[:size]

        # optionally apply func or str formatting to it
        if formatter:
            formatter, *args = formatter
            if isinstance(formatter, str):
                float_format = formatter
            elif callable(formatter):
                # TODO: allow for nan without causing formatter exception
                labels = np.array([formatter(i, *args) for i in labels])
            else:
                raise ToytreeError(
                    "Tuple type label arg must be (feature, Callable, *args) or (feature, formatter).\n"
                    "Examples with Callable:\n"
                    ">>> node_labels=('dist', lambda x: round(x, 2)) \n"
                    ">>> node_labels=('dist', round, 2)\n"
                    ">>> tip_labels=('name', lambda x: f'<i>{x}</i>')"
                    "Examples with format string:\n"
                    ">>> node_labels=('dist', '{:.4g}')"
                )

    # or, user entered a list, tuple, Series, array, etc.
    # which is checked and handled below.

    # float format numeric values, and convert to strings
    flabels = []
    for label in labels:
        try:
            flabel = float_format.format(label)
            flabels.append(str(flabel))
        except (ValueError, TypeError):
            slabel = str(label)
            if slabel == "nan":
                flabels.append("")
            else:
                flabels.append(slabel)

    # validate len and type and convert to an array
    labels = check_arr(flabels, key, size, str)
    return labels


def validate_labels_style(
    tree: ToyTree,
    style: NodeLabelStyle = None,
    **kwargs,
) -> TreeStyle:
    """Ensure tip labels are in px sizes and fill color as ToyColor.

    Parameters
    ----------
    tree:
        A ToyTree.
    style: Mapping
        A dict or SubStyle for Node label style.
    kwargs:
        user style changes.
    """
    if not isinstance(style, NodeLabelStyle):
        style = NodeLabelStyle()

    # update default style with user style args
    for key, val in kwargs.items():
        style[key] = val

    # ensure font-size is in px units
    size = toyplot.units.convert(value=style.font_size, target="px", default="px")
    style.font_size = f"{size:.12g}px"

    # convert fill to a ToyColor
    style.fill = ToyColor(style.fill)
    if style.fill_opacity is None:
        style.fill_opacity = 1.0
    return style


def validate_node_style(
    tree: ToyTree,
    style: NodeStyle = None,
    **kwargs,
) -> Union[NodeStyle, Dict]:
    """Ensure tip labels are in px sizes and fill color as ToyColor.

    Parameters
    ----------
    style: NodeStyle or None
        A NodeStyle object if validating a ToyTree, else None to get a
        default NodeStyle for validating an annotation.
    **kwargs:
        Any arguments accepted by node_style.
    """
    if not isinstance(style, NodeStyle):
        style = NodeStyle()

    # ToyTree drawing validation modifies values in NodeStyle dict.
    # update default style with user style args
    for key, val in kwargs.items():
        style[key] = val

    # convert stroke to ToyColor
    style.stroke = ToyColor(style.stroke)

    # convert fill to ToyColor
    style.fill = ToyColor(style.fill)

    # require opacity settings even though they are None by default
    if style.stroke_opacity is None:
        style.stroke_opacity = 1.0
    if style.fill_opacity is None:
        style.fill_opacity = 1.0

    # return as a NodeStyle object
    return style


def validate_admixture_edges(tree: ToyTree, **kwargs) -> List[Tuple]:
    """Return the aedges list with src,dest expanded to Nodes.
    """
    admixture_edges = kwargs['style'].get("admixture_edges")
    if admixture_edges is None:
        return []

    if not admixture_edges:
        return []

    # ((2,3),(2,4)) or (2,3)
    if isinstance(admixture_edges, tuple):
        if isinstance(admixture_edges[0], tuple):
            admixture_edges = list(admixture_edges)
        else:
            admixture_edges = [admixture_edges]

    # [(2,3),]
    # [((2,3), 4, ...), (...)]
    aedges = []
    for aedge in admixture_edges:
        # user can enter multiple Nodes to get mrca or a single node
        if isinstance(aedge[0], (tuple, list)):
            src = tree.get_mrca_node(*aedge[0]).idx
        else:
            src = tree.get_nodes(aedge[0])[0].idx
        if isinstance(aedge[1], (tuple, list)):
            dst = tree.get_mrca_node(*aedge[1]).idx
        else:
            dst = tree.get_nodes(aedge[1])[0].idx
        aedges.append((src, dst) + tuple(aedge[2:]))
    return aedges


if __name__ == "__main__":

    import toytree

    tree = toytree.rtree.unittree(10, seed=123)

    # node_labels = validate_node_labels(tree, "idx")
    # print(node_labels)

    node_mask = validate_mask(tree, tree.style, {"node_mask": False})
    print(node_mask)

    # node_sizes = validate_node_sizes(tree, False)
    # print(node_sizes)

    # node_labels_style = validate_node_labels_style(tree, style={"fill": "red", "fill-opacity": 0.1}, serialize=True, )
    # print(node_labels_style)

    # node_style = validate_node_style(tree, style={"fill": "red", "fill-opacity": 0.1}, serialize=True)
    # print(node_style)

    # colors = validate_node_colors(tree, ['red'] * tree.nnodes)
    # print(colors)

    # colors = validate_node_colors(tree, ('idx', "Spectral"))
    # print(colors)





#!/usr/bin/env python

"""A replacement for toyplot.html._draw_text() function that allows
for simplifying styles.

This is used in `render_tree` to add tip labels and node labels.
"""

from typing import Dict, Union
from xml.etree import ElementTree as xml
from loguru import logger
import toyplot.style
import toyplot.font
import toyplot.text
from toytree.color.src.concat import concat_style_fix_color

logger = logger.bind(name="toytree")
POP_STYLES = [
    "font-size",
    "font-weight",
    "font-family",
    "vertical-align",
    "white-space",
    "stroke",
]


def render_text(
    root: xml.SubElement,
    text: Union[str, toyplot.text.Layout],
    xpos: float = 0,
    ypos: float = 0,
    style: Dict[str , str] = None,
    angle: float = None,
    title: str = None,
    attributes: Dict[str, str] = None,
) -> None:
    """Add xml.SubElements to the DOM for <text> markers.

    A replacement for toyplot.html._draw_text() function that allows
    for simplifying styles.
    """
    # require helvetica font-family
    style = toyplot.style.combine({"font-family": "helvetica"}, style)
    layout = toyplot.text.layout(text, style, toyplot.font.ReportlabLibrary())

    # DEFAULT STYLE OF LAYOUT. These styles can be updated, but all 
    # have already been set on the <g class=toytree-TipLabels> group
    # and so they are discarded after we get the Layout object.
    # {'fill': '#292724',
    #  'font-family': 'helvetica',
    #  'font-size': '12.0px',
    #  'font-weight': 'normal',
    #  'stroke': 'none',
    #  'vertical-align': 'baseline',
    #  'white-space': 'pre'}

    # a layout has left, right, bottom, height, 
    # layout = (
    #     text if isinstance(text, toyplot.text.Layout) else
    #     toyplot.text.layout(text, style, fonts)
    # )

    # Make a group for the text. e.g., {'class': 'TipLabel'}
    attributes = attributes if attributes is not None else {}
    group = xml.SubElement(root, "g", attrib=attributes)

    # apply a transform to the group
    transform = ""
    if xpos or ypos:
        transform += f"translate({xpos:.6g},{ypos:.6g})"  # %r,%r)" % (x, y)
    if angle:
        transform += f"rotate({-angle:.6g})"  # %r)" % (-angle) # pylint: disable=invalid-unary-operand-type
    if transform:
        group.set("transform", transform)

    # optionally apply a title to the text
    if title is not None:
        xml.SubElement(group, "title").text = str(title)

    # only fill and stroke can differ on an individual text element
    sty = {
        i: style.get(i, None) for i in
        ["fill", "fill-opacity", "stroke", "stroke-opacity"]
    }

    # for each child in layout (element with different style) make xml
    hyperlink = []
    for line in layout.children:
        for box in line.children:

            # draw textbox <text ...>
            if isinstance(box, toyplot.text.TextBox):
                xml.SubElement(
                    group, "text",
                    x=str(box.left),
                    y=str(box.baseline),
                    # style=toyplot.style.to_css(box.style),
                    # style=concat_style_to_str2(sty)
                    style=concat_style_fix_color(sty),
                ).text = box.text

            # NOT TESTED
            elif isinstance(box, toyplot.text.PushHyperlink):
                hyperlink.append(group)
                group = xml.SubElement(
                    group,
                    "a",
                    style=toyplot.style.to_css(box.style),
                )
                group.set("xlink:href", box.href)
                if box.target is not None:
                    group.set("target", box.target)

            elif isinstance(box, toyplot.text.PopHyperlink):
                group = hyperlink.pop()


if __name__ == "__main__":

    root_ = xml.Element('root')
    group_ = xml.SubElement(root_, "g", attrib={"class": "Labels"})    
    render_text(
        root=group_, text="hello", 
        xpos=0, ypos=0, angle=0, 
        attributes={"class": "Label"},
        style={"font-size": "12px", "fill": "red"},
    )
    print(xml.tostring(group_))
